"""Bespoke exceptions and errors."""


class APIResponseError(Exception):
    """APIResponseError exception wrapper to handle API response errors.

    Args:
        Exception : Exception to wrap.
    """


class APIRequestError(Exception):
    """APIRequestError exception wrapper to handle API request erorrs.

    Args:
        Exception : Exception to wrap.
    """


class APIConnectError(Exception):
    """APIConnectError exception wrapper to handle API connection errors.

    Args:
        Exception : Exception to wrap.
    """


class UnrecognizedFormatError(Exception):
    """UnrecognizedFormatError exception wrapper to handle format errors.

    Args:
        Exception : Exception to wrap.
    """


class CredentialError(Exception):
    """CredentialError exception wrapper to handle errors in credentials provided for authentication.

    Args:
        Exception : Exception to wrap.
    """



"""Credentials and token management for Fusion API."""
import datetime
import json
import logging
import os
from datetime import datetime as dt
from typing import Any, Dict, Optional, Tuple
from urllib.parse import urlparse

import jwt
import requests

from . import __version__
from .exceptions import CredentialError

logger = logging.getLogger(__name__)
# Constants analogous to Rust defaults
DEFAULT_GRANT_TYPE = "client_credentials"
DEFAULT_AUTH_URL = "https://authe.jpmorgan.com/as/token.oauth2"
VERSION = __version__


class ProxyType:
    HTTP = "http"
    HTTPS = "https"

    @staticmethod
    def from_str(s: str) -> str:
        if s == "http":
            return ProxyType.HTTP
        elif s == "https":
            return ProxyType.HTTPS
        else:
            raise ValueError("Unrecognized proxy type")


def find_cfg_file(file_path: str) -> str:
    # Attempts to find the config file starting from file_path or its parents
    # until it finds "config/client_credentials.json"
    # This attempts a similar search logic as the Rust code.
    file_path = os.path.abspath(file_path)

    if os.path.isfile(file_path):
        logger.debug(f"Found file at the provided path: {file_path}")
        return file_path

    cwd = os.getcwd()
    cfg_file_name = "client_credentials.json"
    cfg_folder_name = "config"

    start_dir = os.path.dirname(file_path) if os.path.dirname(file_path) else cwd
    start_dir_init = os.path.abspath(start_dir)

    start_dir_abs = os.path.abspath(start_dir)
    while True:
        full_path = os.path.join(start_dir_abs, cfg_folder_name, cfg_file_name)
        if os.path.isfile(full_path):
            logger.debug(f"Found file at: {full_path}")
            return full_path
        parent = os.path.dirname(start_dir_abs)
        if parent == start_dir_abs:
            # Reached root directory
            error_message = (
                f"File {cfg_file_name} not found in {start_dir_init} or any of its parents. "
                f"Current parent: {start_dir}"
            )
            raise FileNotFoundError(error_message)
        start_dir_abs = parent


def fusion_url_to_auth_url(url: str) -> Optional[tuple]:
    logger.debug(f"Trying to form fusion auth url from: {url}")
    try:
        parsed = urlparse(url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError("Invalid URL")
    except ValueError as err:
        raise CredentialError(f"Could not parse URL: {url}") from err
        
    segments = [seg for seg in parsed.path.split("/") if seg]

    # Not a distribution request. No need to authorize
    if "distributions" not in segments:
        logger.debug("Not a distribution request. No need to authorize")
        return None

    # Extract catalog_name
    try:
        cat_idx = segments.index("catalogs")
        catalog_name = segments[cat_idx + 1]
    except (ValueError, IndexError) as err:
        raise CredentialError(
            "'catalogs' segment not found or catalog name missing in the path"
        ) from err

    # Extract dataset_name
    try:
        ds_idx = segments.index("datasets")
        dataset_name = segments[ds_idx + 1]
    except (ValueError, IndexError) as err:
        raise CredentialError(
            "'datasets' segment not found or dataset name missing in the path"
        ) from err

    logger.debug(f"Found Catalog: {catalog_name}, Dataset: {dataset_name}")

    # Reconstruct path until datasets + 1 segment: /catalogs/.../datasets/...
    # Then append /authorize/token
    # segments[:ds_idx+2] gives us everything including the dataset name
    new_path_segments = segments[: ds_idx + 2] + ["authorize", "token"]
    new_path = "/" + "/".join(new_path_segments)

    # Construct the new URL
    # Reuse scheme, netloc from parsed
    fusion_tk_url = f"{parsed.scheme}://{parsed.netloc}{new_path}"
    logger.debug(f"Fusion token URL: {fusion_tk_url}")
    return (fusion_tk_url, catalog_name, dataset_name)


class AuthToken:
    def __init__(self, token: str, expires_in_secs: Optional[int] = None) -> None:
        current_time = int(dt.now(datetime.timezone.utc).timestamp())
        self.token = token
        if expires_in_secs is not None:
            self.expiry = current_time + expires_in_secs
        else:
            self.expiry = None

    def is_expirable(self) -> bool:
        return self.expiry is not None

    def expires_in_secs(self) -> Optional[int]:
        if self.expiry is not None:
            return self.expiry - int(dt.now(datetime.timezone.utc).timestamp())
        return None

    @staticmethod
    def from_token(token: str, expires_in_secs: Optional[int] = None) -> "AuthToken":
        return AuthToken(token, expires_in_secs)

    def as_bearer_header(self) -> Tuple[str, str]:
        return ("Authorization", f"Bearer {self.token}")

    def as_fusion_header(self) -> Tuple[str, str]:
        return ("Fusion-Authorization", f"Bearer {self.token}")


class FusionCredsPersistent:
    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        resource: Optional[str] = None,
        auth_url: Optional[str] = None,
        root_url: Optional[str] = None,
        proxies: Optional[Dict[str, str]] = None,
        grant_type: Optional[str] = None,
        fusion_e2e: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        kid: Optional[str] = None,
        private_key: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.resource = resource
        self.auth_url = auth_url
        self.root_url = root_url
        self.proxies = proxies or {}
        self.grant_type = grant_type or DEFAULT_GRANT_TYPE
        self.fusion_e2e = fusion_e2e
        self.headers = headers or {}
        self.kid = kid
        self.private_key = private_key


class FusionCredentials:
    def __init__(
        self,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        resource: Optional[str] = None,
        auth_url: Optional[str] = None,
        bearer_token: Optional[AuthToken] = None,
        proxies: Optional[Dict[str, str]] = None,
        grant_type: Optional[str] = None,
        fusion_e2e: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        kid: Optional[str] = None,
        private_key: Optional[str] = None,
    ) -> None:

        self.client_id = client_id
        self.client_secret = client_secret
        self.username = username
        self.password = password
        self.resource = resource
        self.auth_url = auth_url or DEFAULT_AUTH_URL
        self.bearer_token = bearer_token
        self.fusion_token = {}
        self.proxies = proxies or {}
        self.grant_type = grant_type or DEFAULT_GRANT_TYPE
        self.fusion_e2e = fusion_e2e
        self.headers = headers or {}
        self.kid = kid
        self.private_key = private_key
        self.http_proxies = self.proxies

    @classmethod
    def from_client_id(
        cls,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        resource: Optional[str] = None,
        auth_url: Optional[str] = None,
        proxies: Optional[Dict[str, str]] = None,  # Added type annotation for proxies
        fusion_e2e: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
        kid: Optional[str] = None,
        private_key: Optional[str] = None,
    ) -> "FusionCredentials":
        return cls(
            client_id=client_id,
            client_secret=client_secret,
            resource=resource,
            auth_url=auth_url or DEFAULT_AUTH_URL,
            proxies=proxies,
            grant_type="client_credentials",
            fusion_e2e=fusion_e2e,
            headers=headers,
            kid=kid,
            private_key=private_key,
        )

    @classmethod
    def from_user_id(
        cls,
        client_id: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        resource: Optional[str] = None,
        auth_url: Optional[str] = None,
        proxies: Optional[Dict] = None,
        fusion_e2e: Optional[bool] = None,
        headers: Optional[Dict] = None,
        kid: Optional[str] = None,
        private_key: Optional[str] = None,
    ) -> "FusionCredentials":
        return cls(
            client_id=client_id,
            username=username,
            password=password,
            resource=resource,
            auth_url=auth_url or DEFAULT_AUTH_URL,
            proxies=proxies,
            grant_type="password",
            fusion_e2e=fusion_e2e,
            headers=headers,
            kid=kid,
            private_key=private_key,
        )

    @classmethod
    def from_bearer_token(
        cls,
        bearer_token: str,
        bearer_token_expiry: Optional[datetime.datetime] = None,
        proxies: Optional[Dict] = None,
        fusion_e2e: Optional[bool] = None,
        headers: Optional[Dict] = None,
    )-> "FusionCredentials":
        if bearer_token is None:
            raise ValueError("Bearer token not provided")
        expiry_secs = None
        if bearer_token_expiry is not None:
            # bearer_token_expiry is a date, we assume midnight expiration
            expiry_dt = datetime.datetime(
                bearer_token_expiry.year,
                bearer_token_expiry.month,
                bearer_token_expiry.day,
            )
            expiry_secs = int(expiry_dt.timestamp()) - int(
                dt.now(datetime.timezone.utc).timestamp()
            )
        return cls(
            bearer_token=AuthToken.from_token(bearer_token, expiry_secs),
            proxies=proxies,
            grant_type="bearer",
            fusion_e2e=fusion_e2e,
            headers=headers,
        )

    @classmethod
    def from_file(cls, file_path: str) -> "FusionCredentials":
        found_path = find_cfg_file(file_path)
        with open(found_path) as f:
            contents = f.read()
        try:
            creds_data = json.loads(contents)
        except json.JSONDecodeError as e:
            raise CredentialError(f"Invalid JSON: {e}\nContents:\n{contents}") from e

        creds = FusionCredsPersistent(**creds_data)

        client_id = creds.client_id or os.environ.get("FUSION_CLIENT_ID")
        if client_id is None:
            raise CredentialError("Missing client ID")

        if creds.grant_type == "client_credentials":
            client_secret = creds.client_secret or os.environ.get(
                "FUSION_CLIENT_SECRET"
            )
            if client_secret is None:
                raise CredentialError("Missing client secret")
            return cls.from_client_id(
                client_id=client_id,
                client_secret=client_secret,
                resource=creds.resource,
                auth_url=creds.auth_url,
                proxies=creds.proxies,
                fusion_e2e=creds.fusion_e2e,
                headers=creds.headers,
                kid=creds.kid,
                private_key=creds.private_key,
            )
        elif creds.grant_type == "bearer":
            return cls.from_bearer_token(
                bearer_token=None,
                bearer_token_expiry=None,
                proxies=creds.proxies,
                fusion_e2e=creds.fusion_e2e,
                headers=creds.headers,
            )
        elif creds.grant_type == "password":
            return cls.from_user_id(
                client_id=client_id,
                username=creds.username,
                password=creds.password,
                resource=creds.resource,
                auth_url=creds.auth_url,
                proxies=creds.proxies,
                fusion_e2e=creds.fusion_e2e,
                headers=creds.headers,
                kid=creds.kid,
                private_key=creds.private_key,
            )
        else:
            raise ValueError("Unrecognized grant type")

    def put_bearer_token(
        self, bearer_token: str, expires_in_secs: Optional[int] = None
    ) -> None:
        self.bearer_token = AuthToken.from_token(bearer_token, expires_in_secs)

    def put_fusion_token(
        self, token_key: str, token: str, expires_in_secs: Optional[int] = None
    ) -> None:
        self.fusion_token[token_key] = AuthToken.from_token(token, expires_in_secs)

    def _refresh_bearer_token(self, force: bool = True, max_remain_secs: int = 30) -> bool:
        if not force and self.bearer_token is not None:
            # Check expiration
            if not self.bearer_token.is_expirable():
                return False
            remain = self.bearer_token.expires_in_secs()
            if remain is not None and remain > max_remain_secs:
                return False

        if self.grant_type == "bearer":
            return True  # Nothing to do

        payload = []
        if self.kid and self.private_key:
            # JWT-based client assertion
            # Construct claims
            iat = int(dt.now(datetime.timezone.utc).timestamp())
            exp = iat + 3600
            claims = {
                "iss": self.client_id or "",
                "aud": self.auth_url or "",
                "sub": self.client_id or "",
                "iat": iat,
                "exp": exp,
                "jti": "id001",
            }
            private_key_jwt = jwt.encode(
                claims, self.private_key, algorithm="RS256", headers={"kid": self.kid}
            )
            payload = [
                ("grant_type", self.grant_type),
                ("client_id", self.client_id or ""),
                (
                    "client_assertion_type",
                    "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
                ),
                (
                    "client_assertion",
                    private_key_jwt
                    if isinstance(private_key_jwt, str)
                    else private_key_jwt.decode("utf-8"),
                ),
                ("resource", self.resource or ""),
            ]
        else:
            if self.grant_type == "client_credentials":
                payload = [
                    ("grant_type", self.grant_type),
                    ("client_id", self.client_id or ""),
                    ("client_secret", self.client_secret or ""),
                    ("aud", self.resource or ""),
                ]
            elif self.grant_type == "password":
                payload = [
                    ("grant_type", self.grant_type),
                    ("client_id", self.client_id or ""),
                    ("username", self.username or ""),
                    ("password", self.password or ""),
                    ("resource", self.resource or ""),
                ]
            else:
                raise ValueError("Unrecognized grant type")

        response = requests.post(
            self.auth_url,
            headers={"User-Agent": f"fusion-python-sdk {VERSION}"},
            data=payload,
            proxies=self.http_proxies,
        )
        if response.status_code != 200:
            raise CredentialError(f"Could not get bearer token: {response.text}")
        res_json = response.json()
        logger.debug(f"Called for Bearer token. Response: {res_json}")

        token = res_json.get("access_token")
        expires_in = res_json.get("expires_in")
        self.put_bearer_token(token, expires_in)
        return True

    def _gen_fusion_token(self, url: str) -> Optional[Tuple[str, Optional[int]]]:
        if not self.bearer_token:
            raise CredentialError("Bearer token is missing")

        headers = {
            "Authorization": f"Bearer {self.bearer_token.token}",
            "User-Agent": f"fusion-python-sdk {VERSION}",
        }

        response = requests.get(url, headers=headers, proxies=self.http_proxies)
        if response.status_code != 200:
            raise CredentialError(f"Error from endpoint: {response.text}")
        res_json = response.json()
        token = res_json.get("access_token")
        expires_in = res_json.get("expires_in")
        logger.debug(f"Got Fusion token, expires in: {expires_in}")
        return token, expires_in

    def get_fusion_token_headers(self, url: str) -> Dict[str, str]:
        # Refresh bearer token if needed
        self._refresh_bearer_token(force=False, max_remain_secs=900)  # 15 mins

        ret = {"User-Agent": f"fusion-python-sdk {VERSION}"}
        if self.fusion_e2e:
            ret["fusion-e2e"] = self.fusion_e2e

        if self.headers:
            for k, v in self.headers.items():
                ret[k] = v

        if not self.bearer_token:
            raise CredentialError("No bearer token set")
        bearer_key, bearer_value = self.bearer_token.as_bearer_header()

        fusion_info = fusion_url_to_auth_url(url)
        if fusion_info is None:
            # Not a distribution request
            ret[bearer_key] = bearer_value
            logger.debug(f"Headers are {ret}")
            return ret

        fusion_tk_url, catalog_name, dataset_name = fusion_info
        token_key = f"{catalog_name}_{dataset_name}"

        # Check existing fusion token
        token_entry = self.fusion_token.get(token_key)
        need_new_token = True
        if token_entry:
            expires_in_secs = token_entry.expires_in_secs()
            if expires_in_secs is not None and expires_in_secs > 900:
                # Enough time remains, no need to refresh
                need_new_token = False

        if need_new_token:
            new_token, new_expires_in = self._gen_fusion_token(fusion_tk_url)
            self.put_fusion_token(token_key, new_token, new_expires_in)
            fusion_key, fusion_value = self.fusion_token[token_key].as_fusion_header()
        else:
            fusion_key, fusion_value = token_entry.as_fusion_header()

        ret[bearer_key] = bearer_value
        ret[fusion_key] = fusion_value

        logger.debug(f"Headers are {ret}")
        return ret

    def get_fusion_token_expires_in(self, token_key: str) -> Optional[int]:
        token = self.fusion_token.get(token_key)
        if token is None:
            return None
        return token.expires_in_secs()



"""Fusion FileSystem."""

import asyncio
import base64
import hashlib
import io
import logging
import time
from copy import deepcopy
from pathlib import Path
from typing import Any, AsyncGenerator, Dict, Generator, List, Optional, Tuple, Union
from urllib.parse import quote, urljoin

import aiohttp
import fsspec
import fsspec.asyn
import pandas as pd
import requests
from fsspec.callbacks import _DEFAULT_CALLBACK
from fsspec.implementations.http import HTTPFile, HTTPFileSystem, sync, sync_wrapper
from fsspec.utils import nullcontext

from .credentials import FusionCredentials
from .utils import cpu_count, get_client, get_default_fs, get_session

logger = logging.getLogger(__name__)
VERBOSE_LVL = 25
DEFAULT_CHUNK_SIZE = 5 * 2 ** 20


class FusionHTTPFileSystem(HTTPFileSystem):  # type: ignore
    """Fusion HTTP filesystem."""

    def __init__(
        self,
        credentials: Optional[
            Union[str, FusionCredentials]
        ] = "config/client_credentials.json",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Same signature as the fsspec HTTPFileSystem.

        Args:
            credentials: Credentials.
            *args: Args.
            **kwargs: Kwargs.
        """

        if "get_client" not in kwargs:
            kwargs["get_client"] = get_client
        if "client_kwargs" not in kwargs:
            if isinstance(credentials, FusionCredentials):
                self.credentials = credentials
            elif isinstance(credentials, str):
                self.credentials = FusionCredentials.from_file(Path(credentials))
            kwargs["client_kwargs"] = {
                "credentials": self.credentials,
                "root_url": "https://fusion.jpmorgan.com/api/v1/",
            }
        elif "client_kwargs" in kwargs and isinstance(
            kwargs["client_kwargs"]["credentials"], FusionCredentials
        ):
            self.credentials = kwargs["client_kwargs"]["credentials"]
        else:
            raise ValueError("Credentials not provided")

        if self.credentials.proxies:
            if "http" in self.credentials.proxies:
                kwargs["proxy"] = self.credentials.proxies["http"]
            elif "https" in self.credentials.proxies:
                kwargs["proxy"] = self.credentials.proxies["https"]

        if "headers" not in kwargs:
            kwargs["headers"] = {"Accept-Encoding": "identity"}
        self.sync_session = get_session(
            self.credentials, kwargs["client_kwargs"].get("root_url")
        )
        super().__init__(*args, **kwargs)

    async def _async_raise_not_found_for_status(self, response: Any, url: str) -> None:
        """Raises FileNotFoundError for 404s, otherwise uses raise_for_status."""
        if response.status == requests.codes.not_found:  # noqa: PLR2004
            self._raise_not_found_for_status(response, url)
        else:
            real_reason = ""
            try:
                real_reason = await response.text()
                response.reason = real_reason
            finally:
                self._raise_not_found_for_status(response, url)

    async def _decorate_url_a(self, url: str) -> str:
        url = (
            urljoin(f'{self.client_kwargs["root_url"]}catalogs/', url)
            if "http" not in url
            else url
        )
        return url

    def _decorate_url(self, url: str) -> str:
        url = (
            urljoin(f'{self.client_kwargs["root_url"]}catalogs/', url)
            if "http" not in url
            else url
        )
        url = url[:-1] if url[-1] == "/" else url
        return url

    async def _isdir(self, path: str) -> bool:
        path = self._decorate_url(path)
        try:
            ret: Dict[str, str] = await self._info(path)
            return ret["type"] == "directory"
        except BaseException:  # noqa: ignore
            logger.exception(VERBOSE_LVL, "Artificial error")
            return False

    async def _changes(self, url: str) -> Dict[Any, Any]:
        url = self._decorate_url(url)
        try:
            session = await self.set_session()
            async with session.get(url, **self.kwargs) as r:
                self._raise_not_found_for_status(r, url)
                try:
                    out: Dict[Any, Any] = await r.json()
                except BaseException:
                    logger.exception(VERBOSE_LVL, f"{url} cannot be parsed to json")
                    out = {}
            return out
        except Exception as ex:
            logger.log(VERBOSE_LVL, f"Artificial error, {ex}")
            raise ex

    async def _ls_real(self, url: str, detail: bool = True, **kwargs: Any) -> Any:
        # ignoring URL-encoded arguments
        clean_url = url
        if "http" not in url:
            url = f'{self.client_kwargs["root_url"]}catalogs/' + url
        kw = self.kwargs.copy()
        kw.update(kwargs)
        session = await self.set_session()
        is_file = False
        size = None
        url = url if url[-1] != "/" else url[:-1]
        url_parts = url.split("/")
        if url_parts[-2] == "distributions":
            async with session.head(
                url + "/operationType/download", **self.kwargs
            ) as r:
                self._raise_not_found_for_status(r, url)
                out = [
                    url.split("/")[6]
                    + "-"
                    + url.split("/")[8]
                    + "-"
                    + url.split("/")[10]
                    + "."
                    + url.split("/")[-1]
                ]

                size = int(r.headers["Content-Length"])
                is_file = True
        else:
            async with session.get(url, **self.kwargs) as r:
                self._raise_not_found_for_status(r, url)
                out = await r.json()

        if not is_file:
            out = [
                urljoin(clean_url + "/", x["identifier"]) for x in out["resources"]
            ]  # type: ignore

        if detail:
            if not is_file:
                return [
                    {
                        "name": u,
                        "size": None,
                        "type": (
                            "directory"
                            if not (u.endswith(("csv", "parquet")))
                            else "file"
                        ),
                    }
                    for u in out
                ]
            else:
                return [{"name": out[0], "size": size, "type": "file"}]
        else:
            return out

    def info(self, path: str, **kwargs: Any) -> Any:
        """Return info.

        Args:
            path: Path.
            **kwargs: Kwargs.

        Returns:

        """
        path = self._decorate_url(path)
        kwargs["keep_protocol"] = True
        res = super().ls(path, detail=True, **kwargs)
        if res[0]["type"] != "file":
            res = super().info(path, **kwargs)
            if path.split("/")[-2] == "datasets":
                target = path.split("/")[-1]
                args = [
                    "/".join(path.split("/")[:-1])
                    + f"/changes?datasets={quote(target)}"
                ]
                res["changes"] = sync(super().loop, self._changes, *args)
        split_path = path.split("/")
        if len(split_path) > 1 and split_path[-2] == "distributions":
            res = res[0]

        return res

    async def _ls(self, url: str, detail: bool = True, **kwargs: Any) -> Any:
        url = await self._decorate_url_a(url)
        return await super()._ls(url, detail, **kwargs)

    def ls(self, url: str, detail: bool = False, **kwargs: Any) -> Any:
        """List resources.

        Args:
            url: Url.
            detail: Detail.
            **kwargs: Kwargs.

        Returns:

        """
        url = self._decorate_url(url)
        ret = super().ls(url, detail=detail, **kwargs)
        keep_protocol = kwargs.pop("keep_protocol", False)
        if detail:
            if not keep_protocol:
                for k in ret:
                    k["name"] = k["name"].split(
                        f'{self.client_kwargs["root_url"]}catalogs/'
                    )[-1]

        elif not keep_protocol:
            return [
                x.split(f'{self.client_kwargs["root_url"]}catalogs/')[-1] for x in ret
            ]

        return ret

    def exists(
        self, url: str, detail: bool = True, **kwargs: Any
    ) -> Any:  # noqa: PLR0913
        """Check existence.

        Args:
            url: Url.
            detail: Detail.
            **kwargs: Kwargs.

        Returns:

        """
        url = self._decorate_url(url)
        return super().exists(url, detail=detail, **kwargs)

    def isfile(self, path: str) -> Union[bool, Any]:
        """Is path a file.

        Args:
            path: Path.

        Returns:

        """
        path = self._decorate_url(path)
        return super().isfile(path)

    def cat(
        self,
        url: str,
        start: Optional[int] = None,
        end: Optional[int] = None,
        **kwargs: Any,
    ) -> Any:
        """Fetch paths' contents.

        Args:
            url: Url.
            start: Start.
            end: End.
            **kwargs: Kwargs.

        Returns:

        """
        url = self._decorate_url(url)
        return super().cat(url, start=start, end=end, **kwargs)

    async def _fetch_range(
        self,
        session: aiohttp.ClientSession,
        url: str,
        start: int,
        end: int,
        output_file: fsspec.spec.AbstractBufferedFile,
    ) -> Any:
        """Fetch a range of bytes from a URL and write it to a file.

        Args:
            url (str): URL to fetch.
            start (int): Start byte.
            end (int): End byte.
            output_file (fsspec.spec.AbstractBufferedFile): File to write to.

        Returns:
            None: None.
        """

        async def fetch() -> None:
            async with session.get(
                url + f"?downloadRange=bytes={start}-{end-1}", **self.kwargs
            ) as response:
                if response.status in [200, 206]:
                    chunk = await response.read()
                    output_file.seek(start)
                    output_file.write(chunk)
                    logger.log(
                        VERBOSE_LVL,
                        f"Wrote {start} - {end} bytes to {output_file.path}"
                    )
                else:
                    response.raise_for_status()

        retries = 5
        for attempt in range(retries):
            try:
                await fetch()
                return
            except Exception as ex:  # noqa: BLE001, PERF203
                if attempt < retries - 1:
                    wait_time = 2 ** attempt  # Exponential backoff
                    logger.log(
                        VERBOSE_LVL,
                        f"Attempt {attempt + 1} failed, retrying in {wait_time} seconds...",  # disable: W1202, C0209
                        exc_info=True,
                    )
                    await asyncio.sleep(wait_time)
                else:
                    logger.log(
                        VERBOSE_LVL,
                        f"Failed to write to {output_file.path}.",
                        exc_info=True,
                    )
                    raise ex

    async def _download_single_file_async(
        self,
        url: str,
        output_file: fsspec.spec.AbstractBufferedFile,
        file_size: int,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        n_threads: int = 10,
    ) -> List[Tuple[bool, str, Optional[str]]]:
        """Download a single file using asynchronous range requests.

        Args:
            url (str): _description_
            output_file (fsspec.spec.AbstractBufferedFile): _description_
            file_size (int): _description_
            chunk_size (int, optional): _description_. Defaults to DEFAULT_CHUNK_SIZE.
            n_threads (int, optional): _description_. Defaults to 10.

        Returns:
            list[tuple[bool, str, Optional[str]]]: Return array.
        """

        coros = []
        session = await self.set_session()
        for start in range(0, file_size, chunk_size):
            end = min(start + chunk_size, file_size)
            task = self._fetch_range(session, url, start, end, output_file)
            coros.append(task)

        # Execute the tasks concurrently
        on_error = "raise"
        out = await fsspec.asyn._run_coros_in_chunks(
            coros, batch_size=n_threads, nofiles=True, return_exceptions=True
        )
        output_file.close()
        if on_error == "raise":
            ex = next(filter(fsspec.asyn.is_exception, out), False)
            if ex:
                return False, output_file.path, str(ex)  # type: ignore

        return True, output_file.path, None  # type: ignore

    def stream_single_file(
        self,
        url: str,
        output_file: fsspec.spec.AbstractBufferedFile,
        block_size: int = DEFAULT_CHUNK_SIZE,
    ) -> Tuple[bool, str, Optional[str]]:
        """Function to stream a single file from the API to a file on disk.

        Args:
            url (str): The URL to call.
            output_file (fsspec.spec.AbstractBufferedFile): The filename handle that the data will be saved into.
            block_size (int, optional): The chunk size to download data. Defaults to DEFAULT_CHUNK_SIZE

        Returns:
            tuple: A tuple

        """

        def get_file() -> None:
            session = self.sync_session
            get_file_kwargs = self.kwargs.copy()
            get_file_kwargs.pop("proxy", None)
            r = session.get(url, **get_file_kwargs)
            r.raise_for_status()
            byte_cnt = 0
            for chunk in r.iter_content(block_size):
                if chunk:
                    byte_cnt += len(chunk)
                    output_file.write(chunk)
            output_file.close()
            logger.log(
                VERBOSE_LVL,
                "Wrote %d bytes to %s",
                byte_cnt,
                getattr(output_file, "name", "unknown"),  # noqa: Q000
            )

        retries = 5
        for attempt in range(retries):
            try:
                get_file()
                return (True, output_file.path, None)  # noqa: UP031
            except Exception as ex:  # noqa: BLE001, PERF203
                if attempt < retries - 1:
                    wait_time = 2 ** attempt  # Exponential backoff
                    logger.log(
                        VERBOSE_LVL,
                        "Attempt %d failed, retrying in %d seconds...",
                        attempt + 1,
                        wait_time,
                        exc_info=True,
                    )
                    time.sleep(wait_time)
                else:
                    output_file.close()
                    logger.log(
                        VERBOSE_LVL,
                        "Failed to write to %s.",
                        output_file.path,
                        exc_info=True,
                    )
                    return False, output_file.path, str(ex)
        return False, output_file.path, None

    def download(
        self,
        lfs: fsspec.AbstractFileSystem,
        rpath: Union[str, Path],
        lpath: Union[str, Path],
        chunk_size: int = 5 * 2 ** 20,
        overwrite: bool = True,
        preserve_original_name: bool = False,
        **kwargs: Any,
    ) -> Any:  # noqa: ignore
        """Download file(s) from remote to local.

        Args:
            lfs (fsspec.AbstractFileSystem): Local filesystem.
            rpath (Union[str, Path]): Remote path.
            lpath (Union[str, Path]): Local path.
            chunk_size (int, optional): Chunk size. Defaults to 5 * 2**20.
            overwrite (bool, optional): True if previously downloaded files should be overwritten. Defaults to True.
            preserve_original_name (bool, optional): True if the original name should be preserved. Defaults to False.
            **kwargs (Any): Kwargs.

        Returns:
            Any: Return value.
        """

        rpath = self._decorate_url(rpath) if isinstance(rpath, str) else rpath
        if not lfs.exists(lpath):
            try:  # noqa: SIM105
                lfs.mkdir(
                    Path(lpath).parent, exist_ok=True, create_parents=True
                )  # noqa: ignore
            except Exception as ex:  # noqa: BLE001, SIM105, F841
                pass

        async def get_headers() -> Any:
            session = await self.set_session()
            async with session.head(rpath, **self.kwargs) as r:
                r.raise_for_status()
                return r.headers

        try:
            headers = sync(self.loop, get_headers)
            if (
                "x-jpmc-file-name" in headers.keys() and preserve_original_name  # noqa: SIM118
            ):  # noqa: SIM118
                file_name = headers.get("x-jpmc-file-name")
                lpath = Path(lpath).parent.joinpath(file_name)
        except Exception as ex:  # noqa: BLE001
            headers = {}
            logger.info(f"Failed to get headers for {rpath}", ex)

        is_local_fs = type(lfs).__name__ == "LocalFileSystem"

        if not overwrite and lfs.exists(lpath):
            return True, lpath, None

        return self.get(
            str(rpath),
            lfs.open(lpath, "wb"),
            chunk_size=chunk_size,
            headers=headers,
            is_local_fs=is_local_fs,
            **kwargs,
        )

    def get(  # disable: W0221
        self,
        rpath: Union[str, io.IOBase],
        lpath: Union[str, io.IOBase],
        chunk_size: int = 5 * 2 ** 20,
        **kwargs: Any,
    ) -> Any:
        """Copy file(s) to local.

        Args:
            rpath: Rpath. Download url.
            lpath: Lpath. Destination path.
            chunk_size: Chunk size.
            **kwargs: Kwargs.

        Returns:

        """

        if isinstance(lpath, str):
            default_fs = get_default_fs()
            lpath = default_fs.open(lpath, "wb")

        rpath = self._decorate_url(rpath) if isinstance(rpath, str) else rpath
        n_threads = cpu_count(is_threading=True)
        file_size = None
        if "headers" in kwargs and "Content-Length" in kwargs["headers"]:
            file_size = int(kwargs["headers"].get("Content-Length"))
        is_local_fs = kwargs.get("is_local_fs", False)
        if n_threads == 1 or not is_local_fs or file_size is None:
            return self.stream_single_file(str(rpath), lpath, block_size=chunk_size)
        else:
            rpath = (
                str(rpath)
                if "operationType/download" in str(rpath)
                else str(rpath) + "/operationType/download"
            )
            return sync(
                self.loop,
                self._download_single_file_async,
                str(rpath),
                lpath,
                file_size,
                chunk_size,
                n_threads,
            )

    @staticmethod
    def _update_kwargs(
        kw: Dict[str, Any],
        headers: Dict[str, str],
        additional_headers: Optional[Dict[str, str]],
    ) -> Dict[str, Any]:
        if "File-Name" in headers:  # noqa: PLR0915
            kw.setdefault("headers", {})
            kw["headers"]["File-Name"] = headers["File-Name"]
        if additional_headers:
            kw["headers"].update(additional_headers)
        return kw

    async def _put_file(  # noqa: PLR0915, PLR0913
        self,
        lpath: Union[str, io.IOBase, fsspec.spec.AbstractBufferedFile],
        rpath: str,
        chunk_size: int = 5 * 2 ** 20,
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        method: str = "post",
        multipart: bool = False,
        additional_headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> None:
        async def put_data() -> AsyncGenerator[Dict[Any, Any], None]:
            # Support passing arbitrary file-like objects
            # and use them instead of streams.
            if isinstance(lpath, io.IOBase):
                context = nullcontext(lpath)
                use_seek = False  # might not support seeking
            else:
                context = open(lpath, "rb")  # noqa: SIM115, PTH123, ASYNC101
                use_seek = True

            with context as f:
                if use_seek:
                    callback.set_size(f.seek(0, 2))
                    f.seek(0)
                else:
                    callback.set_size(getattr(f, "size", None))

                chunk = f.read(chunk_size)
                i = 0
                while chunk:
                    kw = self.kwargs.copy()
                    url = (
                        rpath
                        + f"/operations/upload?operationId={operation_id}&partNumber={i+1}"
                    )
                    kw.update({"headers": kwargs["chunk_headers_lst"][i]})
                    kw = FusionHTTPFileSystem._update_kwargs(
                        kw, headers, additional_headers
                    )
                    async with meth(url=url, data=chunk, **kw) as resp:
                        await self._async_raise_not_found_for_status(resp, rpath)
                        yield await resp.json()
                    i += 1
                    callback.relative_update(len(chunk))
                    chunk = f.read(chunk_size)

        session = await self.set_session()

        method = method.lower()
        if method not in ("post", "put"):
            raise ValueError(
                f"method has to be either 'post' or 'put', not: {method!r}"
            )

        headers = kwargs["headers"]

        meth = getattr(session, method)
        if not multipart:
            kw = self.kwargs.copy()
            kw.update({"headers": headers})
            if additional_headers:
                kw["headers"].update(additional_headers)
            if isinstance(lpath, io.BytesIO):
                lpath.seek(0)
            async with meth(rpath, data=lpath.read(), **kw) as resp:  # type: ignore
                await self._async_raise_not_found_for_status(resp, rpath)
        else:
            kw = self.kwargs.copy()
            kw = FusionHTTPFileSystem._update_kwargs(kw, headers, additional_headers)

            async with session.post(rpath + "/operationType/upload", **kw) as resp:
                await self._async_raise_not_found_for_status(resp, rpath)
                operation_id = await resp.json()

            operation_id = operation_id["operationId"]
            resps = [resp async for resp in put_data()]
            kw = self.kwargs.copy()
            kw.update({"headers": headers})
            kw = FusionHTTPFileSystem._update_kwargs(kw, headers, additional_headers)
            async with session.post(
                url=rpath + f"/operations/upload?operationId={operation_id}",
                json={"parts": resps},
                **kw,
            ) as resp:
                self._raise_not_found_for_status(
                    resp, rpath + f"/operations/upload?operationId={operation_id}"
                )

    @staticmethod
    def _construct_headers(
        file_local: Any,
        dt_from: str,
        dt_to: str,
        dt_created: str,
        chunk_size: int = 5 * 2 ** 20,
        multipart: bool = False,
        file_name: Optional[str] = None,
    ) -> Tuple[Dict[str, str], List[Dict[str, str]]]:
        headers = {
            "Content-Type": "application/octet-stream",
            "x-jpmc-distribution-created-date": dt_created,
            "x-jpmc-distribution-from-date": dt_from,
            "x-jpmc-distribution-to-date": dt_to,
            "Digest": "",  # to be changed to x-jpmc-digest
        }
        if file_name:
            headers["File-Name"] = file_name

        headers["Content-Type"] = (
            "application/json" if multipart else headers["Content-Type"]
        )
        headers_chunks = {"Content-Type": "application/octet-stream", "Digest": ""}

        headers_chunk_lst = []
        hash_sha256 = hashlib.sha256()
        if isinstance(file_local, io.BytesIO):
            file_local.seek(0)
        for chunk in iter(lambda: file_local.read(chunk_size), b""):
            hash_sha256_chunk = hashlib.sha256()
            hash_sha256_chunk.update(chunk)
            hash_sha256.update(hash_sha256_chunk.digest())
            headers_chunks = deepcopy(headers_chunks)
            headers_chunks["Digest"] = (
                "SHA-256=" + base64.b64encode(hash_sha256_chunk.digest()).decode()
            )
            headers_chunk_lst.append(headers_chunks)

        file_local.seek(0)
        if multipart:
            headers["Digest"] = (
                "SHA-256=" + base64.b64encode(hash_sha256.digest()).decode()
            )
        else:
            headers["Digest"] = (
                "SHA-256=" + base64.b64encode(hash_sha256_chunk.digest()).decode()
            )  # noqa: ignore

        return headers, headers_chunk_lst

    def _cloud_copy(  # noqa: PLR0913, PLR0915
        self,
        lpath: Any,
        rpath: Any,
        dt_from: str,
        dt_to: str,
        dt_created: str,
        chunk_size: int = 5 * 2 ** 20,
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        method: str = "put",
        file_name: Optional[str] = None,
        additional_headers: Optional[Dict[str, str]] = None,
    ) -> None:
        async def _get_operation_id(kw: Dict[str, str]) -> Dict[str, Any]:
            session = await self.set_session()
            async with session.post(rpath + "/operationType/upload", **kw) as r:
                await self._async_raise_not_found_for_status(
                    r, rpath + "/operationType/upload"
                )
                res: Dict[str, Any] = await r.json()
                return res

        async def _finish_operation(operation_id: Any, kw: Any) -> None:
            session = await self.set_session()
            async with session.post(
                url=rpath + f"/operations/upload?operationId={operation_id}",
                json={"parts": resps},
                **kw,
            ) as r:
                await self._async_raise_not_found_for_status(
                    r, rpath + f"/operations/upload?operationId={operation_id}"
                )

        def put_data() -> Generator[Dict[str, Any], None, None]:
            async def _meth(url: Any, kw: Any) -> None:
                session = await self.set_session()
                meth = getattr(session, method)
                retry_num = 3
                ex_cnt = 0
                last_ex = None
                while ex_cnt < retry_num:
                    async with meth(url=url, data=chunk, **kw) as resp:
                        try:
                            await self._async_raise_not_found_for_status(resp, url)
                            return await resp.json()  # type: ignore
                        except Exception as ex:  # noqa: BLE001
                            # wait 3 seconds before retrying
                            await asyncio.sleep(3 * (ex_cnt + 1))
                            logger.debug(f"Failed to upload file: {ex}")
                            ex_cnt += 1
                            last_ex = ex

                raise Exception(
                    f"Failed to upload file: {last_ex}, failed after {ex_cnt} exceptions. {last_ex}"
                )

            context = nullcontext(lpath)

            with context as f:
                callback.get_size(getattr(f, "size", None))

            chunk = f.read(chunk_size)
            i = 0
            headers_chunks = {"Content-Type": "application/octet-stream", "Digest": ""}
            while chunk:
                hash_sha256_chunk = hashlib.sha256()
                hash_sha256_chunk.update(chunk)
                hash_sha256_lst[0].update(hash_sha256_chunk.digest())
                headers_chunks = deepcopy(headers_chunks)
                headers_chunks["Digest"] = (
                    "SHA-256=" + base64.b64encode(hash_sha256_chunk.digest()).decode()
                )
                kw = self.kwargs.copy()
                kw.update({"headers": headers_chunks})
                kw = FusionHTTPFileSystem._update_kwargs(
                    kw, headers, additional_headers
                )
                url = (
                    rpath
                    + f"/operations/upload?operationId={operation_id}&partNumber={i+1}"
                )
                yield sync(self.loop, _meth, url, kw)
                i += 1
                callback.relative_update(len(chunk))
                chunk = f.read(chunk_size)

        method = method.lower()
        if method not in ("put", "post"):
            raise ValueError(f"method has to be either 'post' or 'put', not {method!r}")
        hash_sha256 = hashlib.sha256()
        hash_sha256_lst = [hash_sha256]
        headers = {
            "Content-Type": "application/json",
            "x-jpmc-distribution-created-date": dt_created,
            "x-jpmc-distribution-from-date": dt_from,
            "x-jpmc-distribution-to-date": dt_to,
            "Digest": "",  # to be changed to x-jpmc-digest
        }
        if file_name:
            headers["File-Name"] = file_name

        if additional_headers:
            for k, v in additional_headers.items():
                headers[k] = v

        lpath.seek(0)

        kw_op = self.kwargs.copy()
        kw_op = FusionHTTPFileSystem._update_kwargs(kw_op, headers, additional_headers)

        operation_id = sync(self.loop, _get_operation_id, kw_op)["operationId"]
        resps = list(put_data())
        hash_sha256 = hash_sha256_lst[0]
        headers["Digest"] = "SHA-256=" + base64.b64encode(hash_sha256.digest()).decode()
        kw = self.kwargs.copy()
        kw.update({"headers": headers})
        kw = FusionHTTPFileSystem._update_kwargs(kw, headers, additional_headers)
        sync(self.loop, _finish_operation, operation_id, kw)

    def put(  # noqa: PLR0913
        self,
        lpath: str,
        rpath: str,
        chunk_size: int = 5 * 2 ** 20,
        callback: fsspec.callbacks.Callback = _DEFAULT_CALLBACK,
        method: str = "put",
        multipart: bool = False,
        from_date: Optional[str] = None,
        to_date: Optional[str] = None,
        file_name: Optional[str] = None,
        additional_headers: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> Any:
        """Copy file(s) from local.

        Args:
            lpath: Lpath.
            rpath: Rpath.
            chunk_size: Chunk size.
            callback: Callback function.
            method: Method: put/post.
            multipart: Flag which indicated whether it's a multipart uplaod.
            from_date: earliest date of data in upload file
            to_date: latest date of data in upload file
            file_name: Name of the file.
            additional_headers: Additional headers.
            **kwargs: Kwargs.

        Returns:

        """

        if from_date is None or to_date is None:
            dt_from = pd.Timestamp.now().strftime("%Y-%m-%d")
            dt_to = "2199-12-31"
        else:
            dt_from = pd.Timestamp(from_date).strftime("%Y-%m-%d")
            dt_to = pd.Timestamp(to_date).strftime("%Y-%m-%d")

        dt_created = pd.Timestamp.now().strftime("%Y-%m-%d")
        rpath = self._decorate_url(rpath)
        if type(lpath).__name__ in ["S3File"]:
            return self._cloud_copy(
                lpath,
                rpath,
                dt_from,
                dt_to,
                dt_created,
                chunk_size,
                callback,
                method,
                file_name,
                additional_headers,
            )
        headers, chunk_headers_lst = self._construct_headers(
            lpath, dt_from, dt_to, dt_created, chunk_size, multipart, file_name
        )
        kwargs.update({"headers": headers})
        if multipart:
            kwargs.update({"chunk_headers_lst": chunk_headers_lst})
            args = [
                lpath,
                rpath,
                chunk_size,
                callback,
                method,
                multipart,
                additional_headers,
            ]
        else:
            args = [lpath, rpath, None, callback, method, multipart, additional_headers]

        return sync(super().loop, self._put_file, *args, **kwargs)

    def find(
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        **kwargs: Any,
    ) -> Any:
        """Find all file in a folder.

        Args:
            path: Path.
            maxdepth: Max depth.
            withdirs: With dirs, default to False.
            **kwargs: Kwargs.

        Returns:

        """
        path = self._decorate_url(path)
        return super().find(path, maxdepth=maxdepth, withdirs=withdirs, **kwargs)

    def glob(self, path: str, **kwargs: Any) -> Any:
        """Glob.

        Args:
            path: Path.
            **kwargs: Kwargs.

        Returns:

        """

        return super().glob(path, **kwargs)

    def open(
        self, path: str, mode: str = "rb", **kwargs: Any
    ) -> fsspec.spec.AbstractBufferedFile:
        """Open.

        Args:
            path: Path.
            mode: Defaults to rb.
            **kwargs: Kwargs.

        Returns:

        """

        path = self._decorate_url(path)
        return super().open(path, mode, **kwargs)

    def _open(  # noqa: PLR0913
        self,
        path: str,
        mode: str = "rb",
        block_size: Optional[int] = None,
        _autocommit: Optional[bool] = None,
        cache_type: None = None,
        cache_options: None = None,
        size: Optional[None] = None,
        **kwargs: Any,
    ) -> fsspec.spec.AbstractBufferedFile:
        """Make a file-like object.

        Args:
            path (str): Full URL with protocol
            mode (str): must be "rb"
            block_size (int): Bytes to download in one request; use instance value if None. If
            zero, will return a streaming Requests file-like instance.
            autocommit (bool):
            cache_type ():
            cache_options ():
            size ():
            **kwargs ():

        Returns:

        """
        if mode != "rb":
            raise NotImplementedError
        block_size = block_size if block_size is not None else self.block_size
        kw = self.kwargs.copy()
        kw["asynchronous"] = self.asynchronous
        kw.update(kwargs)
        size = size or self.info(path, **kwargs)["size"]
        session = sync(self.loop, self.set_session)
        if block_size and size:
            return FusionFile(
                self,
                path,
                session=session,
                block_size=block_size,
                mode=mode,
                size=size,
                cache_type=cache_type or self.cache_type,
                cache_options=cache_options or self.cache_options,
                loop=self.loop,
                **kw,
            )
        else:
            raise NotImplementedError


class FusionFile(HTTPFile):  # type: ignore
    """Fusion File."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Init."""
        super().__init__(*args, **kwargs)

    async def async_fetch_range(self, start: int, end: int) -> bytes:
        """Download a block of data.

        The expectation is that the server returns only the requested bytes,
        with HTTP code 206. If this is not the case, we first check the headers,
        and then stream the output - if the data size is bigger than we
        requested, an exception is raised.
        """
        logger.debug(f"Fetch range for {self}: {start}-{end}")
        kwargs = self.kwargs.copy()
        headers = kwargs.pop("headers", {}).copy()
        url = self.url + f"/operationType/download?downloadRange=bytes={start}-{end-1}"
        logger.debug(str(url))
        r = await self.session.get(self.fs.encode_url(url), headers=headers, **kwargs)
        async with r:
            if r.status == requests.codes.range_not_satisfiable:  # noqa: PLR2004
                # range request outside file
                return b""
            r.raise_for_status()
            if r.status == requests.codes.partial_content:  # noqa: PLR2004
                # partial content, as expected
                out: bytes = await r.read()
            elif int(r.headers.get("Content-Length", 0)) <= end - start:
                out = await r.read()
            else:
                cl = 0
                out_arr = []
                while True:
                    chunk = await r.content.read(2 ** 20)
                    # data size unknown, let's read until we have enough
                    if chunk:
                        out_arr.append(chunk)
                        cl += len(chunk)
                        if cl > end - start:
                            break
                    else:
                        break
                out = b"".join(out_arr)[: end - start]
            return out

    _fetch_range = sync_wrapper(async_fetch_range)



"""Test dataflow.py"""

import requests
import requests_mock

from fusion.dataflow import InputDataFlow, OutputDataFlow
from fusion.fusion import Fusion


def test_inputdataflow_class_object_representation() -> None:
    """Test the object representation of the Dataflow class."""
    dataflow = InputDataFlow(identifier="my_dataflow", flow_details={"key": "value"})
    assert repr(dataflow)


def test_outputdataflow_class_object_representation() -> None:
    """Test the object representation of the Dataflow class."""
    dataflow = OutputDataFlow(identifier="my_dataflow", flow_details={"key": "value"})
    assert repr(dataflow)


def test_add_registered_attribute(requests_mock: requests_mock.Mocker, fusion_obj: Fusion) -> None:
    """Test the add_registered_attribute method."""
    catalog = "my_catalog"
    dataflow = "TEST_DATAFLOW"
    attribute_identifier = "my_attribute"
    url = f"{fusion_obj.root_url}catalogs/{catalog}/datasets/{dataflow}/attributes/{attribute_identifier}/registration"

    # Mock the POST request to the expected URL with the response JSON
    requests_mock.post(url, json={"isCriticalDataElement": False})

    # Create an InputDataFlow instance and set the client to the Fusion instance
    dataflow_obj = InputDataFlow(identifier="TEST_DATAFLOW")
    dataflow_obj.client = fusion_obj

    # Call the method and verify the response
    resp = dataflow_obj.add_registered_attribute(
        attribute_identifier="my_attribute",
        catalog=catalog,
        return_resp_obj=True
    )

    # Assertions
    assert isinstance(resp, requests.Response)
    status_code = 200
    assert resp.status_code == status_code



import io
import json
from pathlib import Path
from typing import Any, Dict
from unittest import mock
from unittest.mock import MagicMock, patch

import asynctest
import fsspec
import pytest
from aiohttp import ClientResponse
from pytest_mock import MockerFixture

from fusion.credentials import FusionCredentials
from fusion.fusion_filesystem import FusionHTTPFileSystem


@pytest.fixture
def http_fs_instance(credentials_examples: Path) -> FusionHTTPFileSystem:
    """Fixture to create a new instance for each test."""
    creds = FusionCredentials.from_file(credentials_examples)
    return FusionHTTPFileSystem(credentials=creds)


def test_filesystem(
    example_creds_dict: Dict[str, Any], example_creds_dict_https_pxy: Dict[str, Any], tmp_path: Path
) -> None:
    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)
    assert FusionHTTPFileSystem(creds)

    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict_https_pxy, f)
    creds = FusionCredentials.from_file(credentials_file)
    assert FusionHTTPFileSystem(creds)

    kwargs = {"client_kwargs": {"credentials": creds}}

    assert FusionHTTPFileSystem(None, **kwargs)

    kwargs = {"client_kwargs": {"credentials": 3.14}}  # type: ignore
    with pytest.raises(ValueError, match="Credentials not provided"):
        FusionHTTPFileSystem(None, **kwargs)


@pytest.mark.asyncio
async def test_not_found_status(http_fs_instance: FusionHTTPFileSystem) -> None:
    response = asynctest.MagicMock(spec=ClientResponse)
    response.status = 404
    response.text = asynctest.CoroutineMock(return_value="404 NotFound")

    with pytest.raises(FileNotFoundError):
        await http_fs_instance._async_raise_not_found_for_status(response, "http://example.com")

@pytest.mark.asyncio
async def test_other_error_status(credentials: FusionCredentials) -> None:
    response = mock.MagicMock(spec=ClientResponse)
    response.status = 500
    response.text = asynctest.CoroutineMock(return_value="Internal server error")

    instance = FusionHTTPFileSystem(credentials)

    with mock.patch.object(instance, "_raise_not_found_for_status", side_effect=Exception("Custom error")):
        with pytest.raises(Exception, match="Custom error"):
            await instance._async_raise_not_found_for_status(response, "http://example.com")

        response.text.assert_awaited_once()
        assert response.reason == "Internal server error", "The reason should be updated to the response text"


@pytest.mark.asyncio
async def test_successful_status(http_fs_instance: FusionHTTPFileSystem) -> None:
    response = mock.MagicMock(spec=ClientResponse)
    response.status = 200
    response.text = asynctest.CoroutineMock(return_value="Success response body")

    try:
        await http_fs_instance._async_raise_not_found_for_status(response, "http://example.com")
    except FileNotFoundError as e:
        pytest.fail(f"No exception should be raised for a successful response, but got: {e}")


@pytest.mark.asyncio
async def test_decorate_url_with_http_async(http_fs_instance: FusionHTTPFileSystem) -> None:
    url = "resource/path"
    exp_res = f"{http_fs_instance.client_kwargs['root_url']}catalogs/{url}"
    result = await http_fs_instance._decorate_url_a(url)
    assert result == exp_res


@pytest.mark.asyncio
async def test_isdir_true(http_fs_instance: FusionHTTPFileSystem) -> None:
    http_fs_instance._decorate_url = asynctest.CoroutineMock(return_value="decorated_path_dir")  # type: ignore
    http_fs_instance._info = asynctest.CoroutineMock(return_value={"type": "directory"})
    result = await http_fs_instance._isdir("path_dir")
    assert result


@pytest.mark.asyncio
async def test_isdir_false(http_fs_instance: FusionHTTPFileSystem) -> None:
    http_fs_instance._decorate_url = asynctest.CoroutineMock(return_value="decorated_path_file")  # type: ignore
    http_fs_instance._info = asynctest.CoroutineMock(return_value={"type": "file"})
    result = await http_fs_instance._isdir("path_file")
    assert not result

@patch("requests.Session")
def test_stream_single_file(mock_session_class: MagicMock, example_creds_dict: Dict[str, Any], tmp_path: Path) -> None:
    url = "http://example.com/data"
    output_file = MagicMock(spec=fsspec.spec.AbstractBufferedFile)
    output_file.path = "./output_file_path/file.txt"
    output_file.name = "file.txt"

    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)

    # Create a mock response object with the necessary context manager methods
    mock_response = MagicMock()
    mock_response.raise_for_status = MagicMock()
    mock_response.iter_content = MagicMock(return_value=[b"0123456789", b""])
    # Mock the __enter__ method to return the mock response itself
    mock_response.__enter__.return_value = mock_response
    # Mock the __exit__ method to do nothing
    mock_response.__exit__.return_value = None

    # Set up the mock session to return the mock response
    mock_session = MagicMock()
    mock_session.get.return_value = mock_response
    mock_session_class.return_value = mock_session

    # Create an instance of FusionHTTPFileSystem
    http_fs_instance = FusionHTTPFileSystem(credentials=creds)
    http_fs_instance.sync_session = mock_session

    # Run the function
    results = http_fs_instance.stream_single_file(url, output_file)

    # Assertions to verify the behavior
    output_file.write.assert_any_call(b"0123456789")
    assert results == (True, output_file.path, None)


@patch("requests.Session")
def test_stream_single_file_exception(
    mock_session_class: MagicMock, example_creds_dict: Dict[str, Any], tmp_path: Path
) -> None:
    url = "http://example.com/data"
    output_file = MagicMock(spec=fsspec.spec.AbstractBufferedFile)
    output_file.path = "./output_file_path/file.txt"
    output_file.name = "file.txt"

    # Create a mock response object with the necessary context manager methods
    mock_response = MagicMock()
    mock_response.raise_for_status = MagicMock()
    mock_response.iter_content = MagicMock(side_effect=Exception("Test exception"))
    # Mock the __enter__ method to return the mock response itself
    mock_response.__enter__.return_value = mock_response
    # Mock the __exit__ method to do nothing
    mock_response.__exit__.return_value = None

    # Set up the mock session to return the mock response
    mock_session = MagicMock()
    mock_session.get.return_value = mock_response
    mock_session_class.return_value = mock_session

    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)

    # Create an instance of FusionHTTPFileSystem
    http_fs_instance = FusionHTTPFileSystem(credentials=creds)
    http_fs_instance.sync_session = mock_session

    # Run the function and catch the exception
    results = http_fs_instance.stream_single_file(url, output_file)

    # Assertions to verify the behavior
    output_file.close.assert_called_once()
    assert results == (False, output_file.path, "Test exception")

@pytest.mark.asyncio
async def test_download_single_file_async(
    mocker: MockerFixture,  # Add type annotation for mocker
    example_creds_dict: Dict[str, Any], 
    tmp_path: Path
) -> None:
    # Patching the function _run_coros_in_chunks with a mock
    mock_run_coros_in_chunks = mocker.patch("fsspec.asyn._run_coros_in_chunks", new_callable=mocker.AsyncMock)
    
    mock_run_coros_in_chunks.return_value = [True, True, True]

    url = "http://example.com/data"
    output_file = MagicMock(spec=io.IOBase)
    output_file.path = "./output_file_path/file.txt"
    file_size = 20
    chunk_size = 10
    n_threads = 3

    # Mock credentials file
    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)

    # Create the FusionHTTPFileSystem instance
    http_fs_instance = FusionHTTPFileSystem(credentials=creds)
    http_fs_instance.set_session = mocker.AsyncMock(return_value=mocker.AsyncMock())
    http_fs_instance._fetch_range = mocker.AsyncMock()

    # Run the download function
    result = await http_fs_instance._download_single_file_async(url, output_file, file_size, chunk_size, n_threads)

    # Assertions for successful download
    assert result == (True, output_file.path, None)
    output_file.close.assert_called_once()

    # Simulate an exception during the function call
    mock_run_coros_in_chunks.return_value = [Exception("Test exception")]
    result = await http_fs_instance._download_single_file_async(url, output_file, file_size, chunk_size, n_threads)

    # Assertions for failed download
    assert result == (False, output_file.path, "Test exception")
    output_file.close.assert_called()

@pytest.mark.asyncio
async def test_fetch_range_exception(
    mocker: MockerFixture,  # Add type annotation for mocker
    example_creds_dict: Dict[str, Any], 
    tmp_path: Path
) -> None:
    # Mock output file
    output_file = MagicMock(spec=io.IOBase)
    output_file.path = "./output_file_path/file.txt"
    output_file.seek = MagicMock()
    output_file.write = MagicMock()

    # Mock the response with an exception on read
    mock_response = mocker.MagicMock()
    mock_response.raise_for_status = mocker.MagicMock()
    mock_response.read = mocker.MagicMock(side_effect=Exception("Test exception"))
    mock_response.status = 500
    mock_response.__aenter__.return_value = mock_response
    mock_response.__aexit__.return_value = None

    # Mock the ClientSession
    mock_client_session = mocker.MagicMock()
    mock_client_session.get.return_value = mock_response
    mocker.patch("aiohttp.ClientSession", return_value=mock_client_session)

    # Mock credentials file
    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)

    # Create the FusionHTTPFileSystem instance
    http_fs_instance = FusionHTTPFileSystem(credentials=creds)
    http_fs_instance.kwargs = {}

    # Run the function (assuming you're testing a method that handles this exception)
    # You should assert what the function would do in the case of an exception
    await http_fs_instance._fetch_range(mock_client_session, "http://example.com/data", 0, 10, output_file)

    # Assertions
    output_file.seek.assert_not_called()
    output_file.write.assert_not_called()

@pytest.mark.asyncio
async def test_fetch_range_success(
    mocker: MockerFixture,  # Add type annotation for mocker
    example_creds_dict: Dict[str, Any], 
    tmp_path: Path
) -> None:
    url = "http://example.com/data"
    output_file = MagicMock(spec=io.IOBase)
    output_file.path = "./output_file_path/file.txt"
    output_file.seek = MagicMock()
    output_file.write = MagicMock()
    start = 0
    end = 10

    # Mocking the response
    mock_response = mocker.MagicMock()
    mock_response.raise_for_status = mocker.MagicMock()
    mock_response.read = mocker.MagicMock(return_value=b"some data")
    mock_response.status = 200
    mock_response.__aenter__.return_value = mock_response
    mock_response.__aexit__.return_value = None

    # Mocking the session
    mock_session = MagicMock()
    mock_session.get.return_value = mock_response
    mocker.patch("aiohttp.ClientSession", return_value=mock_session)

    # Mocking credentials file
    credentials_file = tmp_path / "client_credentials.json"
    with Path(credentials_file).open("w") as f:
        json.dump(example_creds_dict, f)
    creds = FusionCredentials.from_file(credentials_file)

    # Creating the FusionHTTPFileSystem instance
    http_fs_instance = FusionHTTPFileSystem(credentials=creds)
    http_fs_instance.kwargs = {}

    # Run the function
    await http_fs_instance._fetch_range(mock_session, url, start, end, output_file)

    # Assertions
    output_file.seek.assert_called_once_with(0)
    output_file.write.assert_called_once_with(b"some data")
    mock_response.raise_for_status.assert_not_called()
    mock_session.get.assert_called_once_with(url + f"?downloadRange=bytes={start}-{end-1}", **http_fs_instance.kwargs)


"""Python 3.7 SDK for J.P. Morgan's Fusion platform."""
from __future__ import annotations

import json as js
import logging
import re
import sys
import warnings
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from joblib import Parallel, delayed
from tabulate import tabulate
from tqdm import tqdm

from .attributes import Attribute, Attributes
from .credentials import FusionCredentials
from .dataflow import InputDataFlow, OutputDataFlow
from .dataset import Dataset
from .exceptions import APIResponseError
from .fusion_filesystem import FusionHTTPFileSystem
from .fusion_types import Types
from .product import Product
from .report import Report
from .utils import (
    RECOGNIZED_FORMATS,
    cpu_count,
    distribution_to_filename,
    distribution_to_url,
    get_default_fs,
    get_session,
    is_dataset_raw,
    normalise_dt_param_str,
    path_to_url,
    requests_raise_for_status,
    tqdm_joblib,
    upload_files,
    validate_file_names,
)

if TYPE_CHECKING:
    import fsspec
    import requests


logger = logging.getLogger(__name__)
VERBOSE_LVL = 25


class Fusion:
    """Core Fusion class for API access."""

    @staticmethod
    def _call_for_dataframe(url: str, session: requests.Session) -> pd.DataFrame:
        """Private function that calls an API endpoint and returns the data as a pandas dataframe.

        Args:
            url (str): URL for an API endpoint with valid parameters.
            session (requests.Session): Session object for making the API call.

        Returns:
            pandas.DataFrame: A dataframe containing the requested data.
        """
        response = session.get(url)
        response.raise_for_status()
        table = response.json()["resources"]
        ret_df = pd.DataFrame(table).reset_index(drop=True)
        return ret_df

    @staticmethod
    def _call_for_bytes_object(url: str, session: requests.Session) -> BytesIO:
        """Private function that calls an API endpoint and returns the data as a bytes object in memory.

        Args:
            url (str): URL for an API endpoint with valid parameters.
            session (requests.Session): Session object for making the API call.

        Returns:
            io.BytesIO: In-memory file content.
        """
        response = session.get(url)
        response.raise_for_status()
        return BytesIO(response.content)

    def __init__(
        self,
        credentials: Union[str, dict] = "config/client_credentials.json",
        root_url: str = "https://fusion.jpmorgan.com/api/v1/",
        download_folder: str = "downloads",
        *,
        log_level: int = logging.ERROR,
        log_path: str = ".",
        fs: fsspec.filesystem = None,
    ) -> None:
        """Constructor to instantiate a new Fusion object.

        Args:
            credentials (Union[str, dict]): A path to a credentials file or a dictionary containing credentials.
                Defaults to 'config/client_credentials.json'.
            root_url (str): The API root URL. Defaults to "https://fusion.jpmorgan.com/api/v1/".
            download_folder (str): The folder path where downloaded data files are saved. Defaults to "downloads".
            log_level (int): Set the logging level. Defaults to logging.ERROR.
            log_path (str): The folder path where the log is stored. Defaults to the current directory.
            fs (fsspec.filesystem): filesystem.
        """
        self._default_catalog = "common"

        self.root_url = root_url
        self.download_folder = download_folder
        Path(download_folder).mkdir(parents=True, exist_ok=True)

        if logger.hasHandlers():
            logger.handlers.clear()
        file_handler = logging.FileHandler(filename=f"{log_path}/fusion_sdk.log")
        logging.addLevelName(VERBOSE_LVL, "VERBOSE")
        stdout_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            "%(asctime)s.%(msecs)03d %(name)s:%(levelname)s %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        stdout_handler.setFormatter(formatter)
        logger.addHandler(stdout_handler)
        logger.addHandler(file_handler)
        logger.setLevel(log_level)

        if isinstance(credentials, FusionCredentials):
            self.credentials = credentials
        elif isinstance(credentials, str):
            self.credentials = FusionCredentials.from_file(Path(credentials))  # type: ignore
        else:
            raise ValueError("credentials must be a path to a credentials file or a dictionary")

        self.session = get_session(self.credentials, self.root_url)
        self.fs = fs if fs else get_default_fs()

    def __repr__(self) -> str:
        """Object representation to list all available methods."""
        methods = [
            method_name
            for method_name in dir(Fusion)
            if callable(getattr(Fusion, method_name)) and not method_name.startswith("_")
        ]
        return "Fusion object \nAvailable methods:\n" + tabulate(
            [[method_name] for method_name in methods],
            headers=["Method Name"],
            tablefmt="psql",
        )

    @property
    def default_catalog(self) -> str:
        """Returns the default catalog."""
        return self._default_catalog

    @default_catalog.setter
    def default_catalog(self, catalog: str) -> None:
        """Allow the default catalog, which is "common," to be overridden.

        Args:
            catalog (str): The catalog to use as the default.
        """
        self._default_catalog = catalog

    def _use_catalog(self, catalog: Optional[str]) -> str:
        """Determine which catalog to use in an API call.

        Args:
            catalog (Optional[str]): The catalog value passed as an argument to an API function wrapper.

        Returns:
            str: The catalog to use.
        """
        return catalog if catalog else self.default_catalog

    def get_fusion_filesystem(self) -> FusionHTTPFileSystem:
        """Creates Fusion Filesystem.

        Returns: Fusion Filesystem

        """
        return FusionHTTPFileSystem(client_kwargs={"root_url": self.root_url, "credentials": self.credentials})

    def list_catalogs(self, output: bool = False) -> pd.DataFrame:
        """Lists the catalogs available to the API account.

        Args:
            output (bool, optional): If True then print the dataframe. Defaults to False.

        Returns:
            class:`pandas.DataFrame`: A dataframe with a row for each catalog
        """
        url = f"{self.root_url}catalogs/"
        cat_df = Fusion._call_for_dataframe(url, self.session)

        if output:
            pass

        return cat_df

    def catalog_resources(self, catalog: Optional[str] = None, output: bool = False) -> pd.DataFrame:
        """List the resources contained within the catalog, for example products and datasets.

        Args:
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.

        Returns:
           class:`pandas.DataFrame`: A dataframe with a row for each resource within the catalog
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}"
        cat_df = Fusion._call_for_dataframe(url, self.session)

        if output:
            pass

        return cat_df

    def list_products(
            self,
            contains: Union[str, List[str], None] = None,
            id_contains: bool = False,
            catalog: Optional[str] = None,
            output: bool = False,
            max_results: int = -1,
            display_all_columns: bool = False,
    ) -> pd.DataFrame:
        """Get the products contained in a catalog. A product is a grouping of datasets.

        Args:
            contains (Union[str, list], optional): A string or a list of strings that are product
                identifiers to filter the products list. If a list is provided then it will return
                products whose identifier matches any of the strings. Defaults to None.
            id_contains (bool): Filter datasets only where the string(s) are contained in the identifier,
                ignoring description.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.
            max_results (int, optional): Limit the number of rows returned in the dataframe.
                Defaults to -1 which returns all results.
            display_all_columns (bool, optional): If True displays all columns returned by the API,
                otherwise only the key columns are displayed

        Returns:
            class:`pandas.DataFrame`: a dataframe with a row for each product
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/products"
        full_prod_df: pd.DataFrame = Fusion._call_for_dataframe(url, self.session)

        if contains:
            if isinstance(contains, list):  # noqa: ignore
                contains = "|".join(f"{s}" for s in contains)
            if id_contains:
                filtered_df = full_prod_df[full_prod_df["identifier"].str.contains(contains, case=False)]
            else:
                filtered_df = full_prod_df[
                    full_prod_df["identifier"].str.contains(contains, case=False)
                    | full_prod_df["description"].str.contains(contains, case=False)
                    ]
        else:
            filtered_df = full_prod_df

        filtered_df["category"] = filtered_df.category.str.join(", ")
        filtered_df["region"] = filtered_df.region.str.join(", ")
        if not display_all_columns:
            filtered_df = filtered_df[
                filtered_df.columns.intersection(
                    [
                        "identifier",
                        "title",
                        "region",
                        "category",
                        "status",
                        "description",
                    ]
                )
            ]

        if max_results > -1:
            filtered_df = filtered_df[0:max_results]

        if output:
            pass

        return filtered_df

    def list_datasets(  # noqa: PLR0913
            self,
            contains: Union[str, List[str], None] = None,
            id_contains: bool = False,
            product: Union[str, List[str], None] = None,
            catalog: Optional[str] = None,
            output: bool = False,
            max_results: int = -1,
            display_all_columns: bool = False,
            status: Optional[str] = None,
            dataset_type: Optional[str] = None,
    ) -> pd.DataFrame:
        """Get the datasets contained in a catalog.

        Args:
            contains (Union[str, list], optional): A string or a list of strings that are dataset
                identifiers to filter the datasets list. If a list is provided then it will return
                datasets whose identifier matches any of the strings. Defaults to None.
            id_contains (bool): Filter datasets only where the string(s) are contained in the identifier,
                ignoring description.
            product (Union[str, list], optional): A string or a list of strings that are product
                identifiers to filter the datasets list. Defaults to None.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.
            max_results (int, optional): Limit the number of rows returned in the dataframe.
                Defaults to -1 which returns all results.
            display_all_columns (bool, optional): If True displays all columns returned by the API,
                otherwise only the key columns are displayed
            status (str, optional): filter the datasets by status, default is to show all results.
            dataset_type (str, optional): filter the datasets by type, default is to show all results.

        Returns:
            class:`pandas.DataFrame`: a dataframe with a row for each dataset.
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets"
        ds_df = Fusion._call_for_dataframe(url, self.session)

        if contains:
            if isinstance(contains, list):
                contains = "|".join(f"{s}" for s in contains)
            if id_contains:
                ds_df = ds_df[ds_df["identifier"].str.contains(contains, case=False)]
            else:
                ds_df = ds_df[
                    ds_df["identifier"].str.contains(contains, case=False)
                    | ds_df["description"].str.contains(contains, case=False)
                    ]

        if product:
            url = f"{self.root_url}catalogs/{catalog}/productDatasets"
            prd_df = Fusion._call_for_dataframe(url, self.session)
            prd_df = (
                prd_df[prd_df["product"] == product]
                if isinstance(product, str)
                else prd_df[prd_df["product"].isin(product)]
            )
            ds_df = ds_df[ds_df["identifier"].str.lower().isin(prd_df["dataset"].str.lower())].reset_index(drop=True)

        if max_results > -1:
            ds_df = ds_df[0:max_results]

        ds_df["category"] = ds_df.category.str.join(", ")
        ds_df["region"] = ds_df.region.str.join(", ")
        if not display_all_columns:
            cols = [
                "identifier",
                "title",
                "containerType",
                "region",
                "category",
                "coverageStartDate",
                "coverageEndDate",
                "description",
                "status",
                "type",
            ]
            cols = [c for c in cols if c in ds_df.columns]
            ds_df = ds_df[cols]

        if status is not None:
            ds_df = ds_df[ds_df["status"] == status]

        if dataset_type is not None:
            ds_df = ds_df[ds_df["type"] == dataset_type]

        if output:
            pass

        return ds_df

    def dataset_resources(self, dataset: str, catalog: Optional[str] = None, output: bool = False) -> pd.DataFrame:
        """List the resources available for a dataset, currently this will always be a datasetseries.

        Args:
            dataset (str): A dataset identifier
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.

        Returns:
            class:`pandas.DataFrame`: A dataframe with a row for each resource
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}"
        ds_res_df = Fusion._call_for_dataframe(url, self.session)

        if output:
            pass

        return ds_res_df

    def list_dataset_attributes(
            self,
            dataset: str,
            catalog: Optional[str] = None,
            output: bool = False,
            display_all_columns: bool = False,
    ) -> pd.DataFrame:
        """Returns the list of attributes that are in the dataset.

        Args:
            dataset (str): A dataset identifier
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.
            display_all_columns (bool, optional): If True displays all columns returned by the API,
                otherwise only the key columns are displayed

        Returns:
            class:`pandas.DataFrame`: A dataframe with a row for each attribute
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/attributes"
        ds_attr_df = Fusion._call_for_dataframe(url, self.session)

        if "index" in ds_attr_df.columns:
            ds_attr_df = ds_attr_df.sort_values(by="index").reset_index(drop=True)

        if not display_all_columns:
            ds_attr_df = ds_attr_df[
                ds_attr_df.columns.intersection(
                    [
                        "identifier",
                        "title",
                        "dataType",
                        "isDatasetKey",
                        "description",
                        "source",
                    ]
                )
            ]

        if output:
            pass

        return ds_attr_df

    def list_datasetmembers(
            self,
            dataset: str,
            catalog: Optional[str] = None,
            output: bool = False,
            max_results: int = -1,
    ) -> pd.DataFrame:
        """List the available members in the dataset series.

        Args:
            dataset (str): A dataset identifier
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.
            max_results (int, optional): Limit the number of rows returned in the dataframe.
                Defaults to -1 which returns all results.

        Returns:
            class:`pandas.DataFrame`: a dataframe with a row for each dataset member.
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/datasetseries"
        ds_members_df = Fusion._call_for_dataframe(url, self.session)

        if max_results > -1:
            ds_members_df = ds_members_df[0:max_results]

        if output:
            pass

        return ds_members_df

    def datasetmember_resources(
            self,
            dataset: str,
            series: str,
            catalog: Optional[str] = None,
            output: bool = False,
    ) -> pd.DataFrame:
        """List the available resources for a datasetseries member.

        Args:
            dataset (str): A dataset identifier
            series (str): The datasetseries identifier
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.

        Returns:
            class:`pandas.DataFrame`: A dataframe with a row for each datasetseries member resource.
                Currently, this will always be distributions.
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/datasetseries/{series}"
        ds_mem_res_df = Fusion._call_for_dataframe(url, self.session)

        if output:
            pass

        return ds_mem_res_df

    def list_distributions(
            self,
            dataset: str,
            series: str,
            catalog: Optional[str] = None,
            output: bool = False,
    ) -> pd.DataFrame:
        """List the available distributions (downloadable instances of the dataset with a format type).

        Args:
            dataset (str): A dataset identifier
            series (str): The datasetseries identifier
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True then print the dataframe. Defaults to False.

        Returns:
            class:`pandas.DataFrame`: A dataframe with a row for each distribution.
        """
        catalog = self._use_catalog(catalog)

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/datasetseries/{series}/distributions"
        distros_df = Fusion._call_for_dataframe(url, self.session)

        if output:
            pass

        return distros_df

    def _resolve_distro_tuples(
            self,
            dataset: str,
            dt_str: str = "latest",
            dataset_format: str = "parquet",
            catalog: Optional[str] = None,
    ) -> list:
        """Resolve distribution tuples given specification params.

        A private utility function to generate a list of distribution tuples.
        Each tuple is a distribution, identified by catalog, dataset id,
        datasetseries member id, and the file format.

        Args:
            dataset (str): A dataset identifier
            dt_str (str, optional): Either a single date or a range identified by a start or end date,
                or both separated with a ":". Defaults to 'latest' which will return the most recent
                instance of the dataset.
            dataset_format (str, optional): The file format, e.g. CSV or Parquet. Defaults to 'parquet'.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.

        Returns:
            list: a list of tuples, one for each distribution
        """
        catalog = self._use_catalog(catalog)

        datasetseries_list = self.list_datasetmembers(dataset, catalog)
        if len(datasetseries_list) == 0:
            raise AssertionError(f"There are no dataset members for dataset {dataset} in catalog {catalog}")

        if datasetseries_list.empty:
            raise APIResponseError(  # pragma: no cover
                f"No data available for dataset {dataset}. "
                f"Check that a valid dataset identifier and date/date range has been set."
            )

        if dt_str == "latest":
            dt_str = (
                datasetseries_list[
                    datasetseries_list["createdDate"] == datasetseries_list["createdDate"].to_numpy().max()
                    ]
                .sort_values(by="identifier")
                .iloc[-1]["identifier"]
            )
            datasetseries_list = datasetseries_list[datasetseries_list["identifier"] == dt_str]
        else:
            parsed_dates = normalise_dt_param_str(dt_str)
            if len(parsed_dates) == 1:
                parsed_dates = (parsed_dates[0], parsed_dates[0])

            if parsed_dates[0]:
                datasetseries_list = datasetseries_list[
                    pd.Series([pd.to_datetime(i, errors="coerce") for i in datasetseries_list["identifier"]])
                    >= pd.to_datetime(parsed_dates[0])
                    ].reset_index()

            if parsed_dates[1]:
                datasetseries_list = datasetseries_list[
                    pd.Series([pd.to_datetime(i, errors="coerce") for i in datasetseries_list["identifier"]])
                    <= pd.to_datetime(parsed_dates[1])
                    ].reset_index()

        if len(datasetseries_list) == 0:
            raise APIResponseError(  # pragma: no cover
                f"No data available for dataset {dataset} in catalog {catalog}.\n"
                f"Check that a valid dataset identifier and date/date range has been set."
            )

        required_series = list(datasetseries_list["@id"])
        tups = [(catalog, dataset, series, dataset_format) for series in required_series]

        return tups

    def download(  # noqa: PLR0912, PLR0913
            self,
            dataset: str,
            dt_str: str = "latest",
            dataset_format: str = "parquet",
            catalog: Optional[str] = None,
            n_par: Optional[int] = None,
            show_progress: bool = True,
            force_download: bool = False,
            download_folder: Optional[str] = None,
            return_paths: bool = False,
            partitioning: Optional[str] = None,
            preserve_original_name: bool = False,
    ) -> Optional[List[Tuple[bool, str, Optional[str]]]]:
        """Downloads the requested distributions of a dataset to disk.

        Args:
            dataset (str): A dataset identifier
            dt_str (str, optional): Either a single date or a range identified by a start or end date,
                or both separated with a ":". Defaults to 'latest' which will return the most recent
                instance of the dataset. If more than one series member exists on the latest date, the
                series member identifiers will be sorted alphabetically and the last one will be downloaded.
            dataset_format (str, optional): The file format, e.g. CSV or Parquet. Defaults to 'parquet'.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            n_par (int, optional): Specify how many distributions to download in parallel.
                Defaults to all cpus available.
            show_progress (bool, optional): Display a progress bar during data download Defaults to True.
            force_download (bool, optional): If True then will always download a file even
                if it is already on disk. Defaults to True.
            download_folder (str, optional): The path, absolute or relative, where downloaded files are saved.
                Defaults to download_folder as set in __init__
            return_paths (bool, optional): Return paths and success statuses of the downloaded files.
            partitioning (str, optional): Partitioning specification.
            preserve_original_name (bool, optional): Preserve the original name of the file. Defaults to False.

        Returns:

        """
        catalog = self._use_catalog(catalog)

        valid_date_range = re.compile(r"^(\d{4}\d{2}\d{2})$|^((\d{4}\d{2}\d{2})?([:])(\d{4}\d{2}\d{2})?)$")

        if valid_date_range.match(dt_str) or dt_str == "latest":
            required_series = self._resolve_distro_tuples(dataset, dt_str, dataset_format, catalog)
        else:
            # sample data is limited to csv
            if dt_str == "sample":
                dataset_format = self.list_distributions(dataset, dt_str, catalog)["identifier"].iloc[0]
            required_series = [(catalog, dataset, dt_str, dataset_format)]

        if dataset_format not in RECOGNIZED_FORMATS + ["raw"]:
            raise ValueError(f"Dataset format {dataset_format} is not supported")

        if not download_folder:
            download_folder = self.download_folder

        download_folders = [download_folder] * len(required_series)

        if partitioning == "hive":
            members = [series[2].strip("/") for series in required_series]
            download_folders = [
                f"{download_folders[i]}/{series[0]}/{series[1]}/{members[i]}"
                for i, series in enumerate(required_series)
            ]

        for d in download_folders:
            if not self.fs.exists(d):
                self.fs.mkdir(d, create_parents=True)

        n_par = cpu_count(n_par)
        download_spec = [
            {
                "lfs": self.fs,
                "rpath": distribution_to_url(
                    self.root_url,
                    series[1],
                    series[2],
                    series[3],
                    series[0],
                    is_download=True,
                ),
                "lpath": distribution_to_filename(
                    download_folders[i],
                    series[1],
                    series[2],
                    series[3],
                    series[0],
                    partitioning=partitioning,
                ),
                "overwrite": force_download,
                "preserve_original_name": preserve_original_name,
            }
            for i, series in enumerate(required_series)
        ]

        logger.log(
            VERBOSE_LVL,
            f"Beginning {len(download_spec)} downloads in batches of {n_par}",
        )
        if show_progress:
            with tqdm_joblib(tqdm(total=len(download_spec))):
                res = Parallel(n_jobs=n_par)(
                    delayed(self.get_fusion_filesystem().download)(**spec) for spec in download_spec
                )
        else:
            res = Parallel(n_jobs=n_par)(
                delayed(self.get_fusion_filesystem().download)(**spec) for spec in download_spec
            )

        if (len(res) > 0) and (not all(r[0] for r in res)):
            for r in res:
                if not r[0]:
                    warnings.warn(f"The download of {r[1]} was not successful", stacklevel=2)
        return res if return_paths else None

    def to_bytes(
            self,
            dataset: str,
            series_member: str,
            dataset_format: str = "parquet",
            catalog: Optional[str] = None,
    ) -> BytesIO:
        """Returns an instance of dataset (the distribution) as a bytes object.

        Args:
            dataset (str): A dataset identifier
            series_member (str,): A dataset series member identifier
            dataset_format (str, optional): The file format, e.g. CSV or Parquet. Defaults to 'parquet'.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
        """

        catalog = self._use_catalog(catalog)

        url = distribution_to_url(
            self.root_url,
            dataset,
            series_member,
            dataset_format,
            catalog,
        )

        return Fusion._call_for_bytes_object(url, self.session)

    def upload(  # noqa: PLR0913
            self,
            path: str,
            dataset: Optional[str] = None,
            dt_str: str = "latest",
            catalog:Optional[str] = None,
            n_par: Optional[int] = None,
            show_progress: bool = True,
            return_paths: bool = False,
            multipart: bool = True,
            chunk_size: int = 5 * 2 ** 20,
            from_date: Optional[str] = None,
            to_date: Optional[str] = None,
            preserve_original_name: Optional[bool] = False,
            additional_headers: Optional[Dict[str, str]] = None,
    ) -> Optional[List[Tuple[bool, str, Optional[str]]]]:
        """Uploads the requested files/files to Fusion.

        Args:
            path (str): path to a file or a folder with files
            dataset (str, optional): Dataset identifier to which the file will be uploaded (for single file only).
                                    If not provided the dataset will be implied from file's name.
            dt_str (str, optional): A file name. Can be any string but is usually a date.
                                    Defaults to 'latest' which will return the most recent.
                                    Relevant for a single file upload only. If not provided the dataset will
                                    be implied from file's name.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            n_par (int, optional): Specify how many distributions to download in parallel.
                Defaults to all cpus available.
            show_progress (bool, optional): Display a progress bar during data download Defaults to True.
            return_paths (bool, optional): Return paths and success statuses of the downloaded files.
            multipart (bool, optional): Is multipart upload.
            chunk_size (int, optional): Maximum chunk size.
            from_date (str, optional): start of the data date range contained in the distribution,
                defaults to upoad date
            to_date (str, optional): end of the data date range contained in the distribution,
                defaults to upload date.
            preserve_original_name (bool, optional): Preserve the original name of the file. Defaults to False.

        Returns:


        """
        catalog = self._use_catalog(catalog)

        if not self.fs.exists(path):
            raise RuntimeError("The provided path does not exist")

        fs_fusion = self.get_fusion_filesystem()
        if self.fs.info(path)["type"] == "directory":
            file_path_lst = self.fs.find(path)
            local_file_validation = validate_file_names(file_path_lst, fs_fusion)
            file_path_lst = [f for flag, f in zip(local_file_validation, file_path_lst) if flag]
            file_name = [f.split("/")[-1] for f in file_path_lst]
            is_raw_lst = is_dataset_raw(file_path_lst, fs_fusion)
            local_url_eqiv = [path_to_url(i, r) for i, r in zip(file_path_lst, is_raw_lst)]
        else:
            file_path_lst = [path]
            if not catalog or not dataset:
                local_file_validation = validate_file_names(file_path_lst, fs_fusion)
                file_path_lst = [f for flag, f in zip(local_file_validation, file_path_lst) if flag]
                is_raw_lst = is_dataset_raw(file_path_lst, fs_fusion)
                local_url_eqiv = [path_to_url(i, r) for i, r in zip(file_path_lst, is_raw_lst)]
                if preserve_original_name:
                    raise ValueError("preserve_original_name can only be used when catalog and dataset are provided.")
            else:
                date_identifier = re.compile(r"^(\d{4})(\d{2})(\d{2})$")
                if date_identifier.match(dt_str):
                    dt_str = dt_str if dt_str != "latest" else pd.Timestamp("today").date().strftime("%Y%m%d")
                    dt_str = pd.Timestamp(dt_str).date().strftime("%Y%m%d")

                if catalog not in fs_fusion.ls("") or dataset not in [
                    i.split("/")[-1] for i in fs_fusion.ls(f"{catalog}/datasets")
                ]:
                    msg = (
                        f"File file has not been uploaded, one of the catalog: {catalog} "
                        f"or dataset: {dataset} does not exit."
                    )
                    warnings.warn(msg, stacklevel=2)
                    return [(False, path, msg)]
                file_format = path.split(".")[-1]
                file_name = [path.split("/")[-1]]
                file_format = "raw" if file_format not in RECOGNIZED_FORMATS else file_format

                local_url_eqiv = [
                    "/".join(distribution_to_url("", dataset, dt_str, file_format, catalog, False).split("/")[1:])
                ]

        if not preserve_original_name:
            data_map_df = pd.DataFrame([file_path_lst, local_url_eqiv]).T
            data_map_df.columns = pd.Index(["path", "url"])
        else:
            data_map_df = pd.DataFrame([file_path_lst, local_url_eqiv, file_name]).T
            data_map_df.columns = pd.Index(["path", "url", "file_name"])

        n_par = cpu_count(n_par)
        parallel = len(data_map_df) > 1
        res = upload_files(
            fs_fusion,
            self.fs,
            data_map_df,
            parallel=parallel,
            n_par=n_par,
            multipart=multipart,
            chunk_size=chunk_size,
            show_progress=show_progress,
            from_date=from_date,
            to_date=to_date,
            additional_headers=additional_headers,
        )

        if not all(r[0] for r in res):
            failed_res = [r for r in res if not r[0]]
            msg = f"Not all uploads were successfully completed. The following failed:\n{failed_res}"
            logger.warning(msg)
            warnings.warn(msg, stacklevel=2)

        return res if return_paths else None

    def from_bytes(  # noqa: PLR0913
            self,
            data: BytesIO,
            dataset: str,
            series_member: str = "latest",
            catalog: Optional[str] = None,
            distribution: str = "parquet",
            show_progress: bool = True,
            return_paths: bool = False,
            chunk_size: int = 5 * 2 ** 20,
            from_date: Optional[str] = None,
            to_date: Optional[str] = None,
            file_name: Optional[str] = None,
            **kwargs: Any,  # noqa: ARG002
    ) -> Optional[List[Tuple[bool, str, Optional[str]]]]:
        """Uploads data from an object in memory.

        Args:
            data (str): an object in memory to upload
            dataset (str): Dataset name to which the bytes will be uploaded.
            series_member (str, optional): A single date or label. Defaults to 'latest' which will return
                the most recent.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            distribution (str, optional): A distribution type, e.g. a file format or raw
            show_progress (bool, optional): Display a progress bar during data download Defaults to True.
            return_paths (bool, optional): Return paths and success statuses of the downloaded files.
            chunk_size (int, optional): Maximum chunk size.
            from_date (str, optional): start of the data date range contained in the distribution,
                defaults to upload date
            to_date (str, optional): end of the data date range contained in the distribution, defaults to upload date.
            file_name (str, optional): file name to be used for the uploaded file. Defaults to Fusion standard naming.

        Returns:
            Optional[list[tuple[bool, str, Optional[str]]]: a list of tuples, one for each distribution

        """
        catalog = self._use_catalog(catalog)

        fs_fusion = self.get_fusion_filesystem()
        if distribution not in RECOGNIZED_FORMATS + ["raw"]:
            raise ValueError(f"Dataset format {distribution} is not supported")

        is_raw = js.loads(fs_fusion.cat(f"{catalog}/datasets/{dataset}"))["isRawData"]
        local_url_eqiv = path_to_url(f"{dataset}__{catalog}__{series_member}.{distribution}", is_raw)

        data_map_df = pd.DataFrame(["", local_url_eqiv, file_name]).T
        data_map_df.columns = ["path", "url", "file_name"]  # type: ignore

        res = upload_files(
            fs_fusion,
            data,
            data_map_df,
            parallel=False,
            n_par=1,
            multipart=False,
            chunk_size=chunk_size,
            show_progress=show_progress,
            from_date=from_date,
            to_date=to_date,
        )

        if not all(r[0] for r in res):
            failed_res = [r for r in res if not r[0]]
            msg = f"Not all uploads were successfully completed. The following failed:\n{failed_res}"
            logger.warning(msg)
            warnings.warn(msg, stacklevel=2)

        return res if return_paths else None

    def list_dataset_lineage(
        self,
        dataset_id: str,
        catalog: Optional[str] = None,
        output: bool = False,
        max_results: int = -1,
    ) -> pd.DataFrame:
        """List the upstream and downstream lineage of the dataset.

        Args:
            dataset_id (str): A dataset identifier.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            output (bool, optional): If True, then print the dataframe. Defaults to False.
            max_results (int, optional): Limit the number of rows returned in the dataframe.
                Defaults to -1, which returns all results.

        Returns:
            pd.DataFrame: A dataframe with a row for each resource.

        Raises:
            HTTPError: If the dataset is not found in the catalog.
        """
        catalog = self._use_catalog(catalog)

        url_dataset = f"{self.root_url}catalogs/{catalog}/datasets/{dataset_id}"
        resp_dataset = self.session.get(url_dataset)
        resp_dataset.raise_for_status()

        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset_id}/lineage"
        resp = self.session.get(url)
        data = resp.json()
        relations_data = data["relations"]

        restricted_datasets = [
            dataset_metadata["identifier"]
            for dataset_metadata in data["datasets"]
            if dataset_metadata.get("status", None) == "Restricted"
        ]

        data_dict = {}

        for entry in relations_data:
            source_dataset_id = entry["source"]["dataset"]
            source_catalog = entry["source"]["catalog"]
            destination_dataset_id = entry["destination"]["dataset"]
            destination_catalog = entry["destination"]["catalog"]

            if destination_dataset_id == dataset_id:
                for dataset in data["datasets"]:
                    if dataset["identifier"] == source_dataset_id and dataset.get("status", None) != "Restricted":
                        source_dataset_title = dataset["title"]
                    elif dataset["identifier"] == source_dataset_id and dataset.get("status", None) == "Restricted":
                        source_dataset_title = "Access Restricted"
                data_dict[source_dataset_id] = (
                    "source",
                    source_catalog,
                    source_dataset_title,
                )

            if source_dataset_id == dataset_id:
                for dataset in data["datasets"]:
                    if dataset["identifier"] == destination_dataset_id and dataset.get("status", None) != "Restricted":
                        destination_dataset_title = dataset["title"]
                    elif (
                        dataset["identifier"] == destination_dataset_id and dataset.get("status", None) == "Restricted"
                    ):
                        destination_dataset_title = "Access Restricted"
                data_dict[destination_dataset_id] = (
                    "produced",
                    destination_catalog,
                    destination_dataset_title,
                )

        output_data = {
            "type": [v[0] for v in data_dict.values()],
            "dataset_identifier": list(data_dict.keys()),
            "title": [v[2] for v in data_dict.values()],
            "catalog": [v[1] for v in data_dict.values()],
        }

        lineage_df = pd.DataFrame(output_data)
        lineage_df.loc[
            lineage_df["dataset_identifier"].isin(restricted_datasets),
            ["dataset_identifier", "catalog", "title"],
        ] = "Access Restricted"

        if max_results > -1:
            lineage_df = lineage_df[0:max_results]

        if output:
            print(lineage_df)

        return lineage_df


    def create_dataset_lineage(
            self: Fusion,
            base_dataset: str,
            source_dataset_catalog_mapping: Union[pd.DataFrame, List[Dict[str, str]]],
            catalog: Optional[str] = None,
            return_resp_obj: bool = False,
        ) -> Optional[requests.Response]:
        """Upload lineage to a dataset.

        Args:
            base_dataset (str): A dataset identifier to which you want to add lineage.
            source_dataset_catalog_mapping (Union[pd.DataFrame, list[dict[str]]]): Mapping for the dataset
                identifier(s) and catalog(s) from which to add lineage.
            catalog (Optional[str], optional): Catalog identifier. Defaults to None.
            return_resp_obj (bool, optional): If True, then return the response object. Defaults to False.

        Raises:
            ValueError: If source_dataset_catalog_mapping is not a pandas DataFrame or a list of dictionaries.
            HTTPError: If the request is unsuccessful.

        Examples:
            Creating lineage from a pandas DataFrame.
            >>> data = [{"dataset": "a", "catalog": "a"}, {"dataset": "b", "catalog": "b"}]
            >>> df = pd.DataFrame(data)
            >>> fusion = Fusion()
            >>> fusion.create_dataset_lineage(base_dataset="c", source_dataset_catalog_mapping=df, catalog="c")

            Creating lineage from a list of dictionaries.
            >>> data = [{"dataset": "a", "catalog": "a"}, {"dataset": "b", "catalog": "b"}]
            >>> fusion = Fusion()
            >>> fusion.create_dataset_lineage(base_dataset="c", source_dataset_catalog_mapping=data, catalog="c")

        """
        catalog = self._use_catalog(catalog)

        if isinstance(source_dataset_catalog_mapping, pd.DataFrame):
            dataset_mapping_list = [
                {"dataset": row["dataset"], "catalog": row["catalog"]}
                for _, row in source_dataset_catalog_mapping.iterrows()
            ]
        elif isinstance(source_dataset_catalog_mapping, list):
            dataset_mapping_list = source_dataset_catalog_mapping
        else:
            raise ValueError("source_dataset_catalog_mapping must be a pandas DataFrame or a list of dictionaries.")

        data = {"source": dataset_mapping_list}

        url = f"{self.root_url}catalogs/{catalog}/datasets/{base_dataset}/lineage"

        resp = self.session.post(url, json=data)
        resp.raise_for_status()

        return resp if return_resp_obj else None

    def list_product_dataset_mapping(
            self: Fusion,
            dataset: Union[str, List[str], None] = None,
            product: Union[str, List[str], None] = None,
            catalog: Union[str, None] = None,
        ) -> pd.DataFrame:
            """Get the product to dataset linking contained in a catalog. A product is a grouping of datasets.

            Args:
                dataset (Union[str, List[str], None], optional): A string or list of strings that are dataset
                identifiers to filter the output. If a list is provided then it will return
                datasets whose identifier matches any of the strings. Defaults to None.
                product (Union[str, List[str], None], optional): A string or list of strings that are product
                identifiers to filter the output. If a list is provided then it will return
                products whose identifier matches any of the strings. Defaults to None.
                catalog (Union[str, None], optional): A catalog identifier. Defaults to 'common'.

            Returns:
                pd.DataFrame: A dataframe with a row for each dataset to product mapping.
            """
            catalog = self._use_catalog(catalog)
            url = f"{self.root_url}catalogs/{catalog}/productDatasets"
            mapping_df = pd.DataFrame(self._call_for_dataframe(url, self.session))

            if dataset:
                if isinstance(dataset, list):
                    contains = "|".join(f"{s}" for s in dataset)
                    mapping_df = mapping_df[mapping_df["dataset"].str.contains(contains, case=False)]
                if isinstance(dataset, str):
                    mapping_df = mapping_df[mapping_df["dataset"].str.contains(dataset, case=False)]
            if product:
                if isinstance(product, list):
                    contains = "|".join(f"{s}" for s in product)
                    mapping_df = mapping_df[mapping_df["product"].str.contains(contains, case=False)]
                if isinstance(product, str):
                    mapping_df = mapping_df[mapping_df["product"].str.contains(product, case=False)]
            return mapping_df

    def product(
            self: Fusion,
            identifier: str,
            title: str = "",
            category: Union[str, List[str], None] = None,
            short_abstract: str = "",
            description: str = "",
            is_active: bool = True,
            is_restricted: Union[bool, None] = None,
            maintainer: Union[str, List[str], None] = None,
            region: Union[str, List[str]] = "Global",
            publisher: str = "J.P. Morgan",
            sub_category: Union[str, List[str], None] = None,
            tag: Union[str, List[str], None] = None,
            delivery_channel: Union[str, List[str]] = "API",
            theme: Union[str, None] = None,
            release_date: Union[str, None] = None,
            language: str = "English",
            status: str = "Available",
            image: str = "",
            logo: str = "",
            dataset: Union[str, List[str], None] = None,
            **kwargs: Any,
        ) -> Product:
            """Instantiate a Product object with this client for metadata creation."""
            product_obj = Product(
                identifier=identifier,
                title=title,
                category=category,
                short_abstract=short_abstract,
                description=description,
                is_active=is_active,
                is_restricted=is_restricted,
                maintainer=maintainer,
                region=region,
                publisher=publisher,
                sub_category=sub_category,
                tag=tag,
                delivery_channel=delivery_channel,
                theme=theme,
                release_date=release_date,
                language=language,
                status=status,
                image=image,
                logo=logo,
                dataset=dataset,
                **kwargs,
            )
            product_obj.client = self
            return product_obj

    def dataset(
            self: Fusion,
            identifier: str,
            title: str = "",
            category: Union[str, List[str], None] = None,
            description: str = "",
            frequency: str = "Once",
            is_internal_only_dataset: bool = False,
            is_third_party_data: bool = True,
            is_restricted: Union[bool, None] = None,
            is_raw_data: bool = True,
            maintainer: Union[str, None] = "J.P. Morgan Fusion",
            source: Union[str, List[str], None] = None,
            region: Union[str, List[str], None] = None,
            publisher: str = "J.P. Morgan",
            product: Union[str, List[str], None] = None,
            sub_category: Union[str, List[str], None] = None,
            tags: Union[str, List[str], None] = None,
            created_date: Union[str, None] = None,
            modified_date: Union[str, None] = None,
            delivery_channel: Union[str, List[str]] = "API",
            language: str = "English",
            status: str = "Available",
            type_: Union[str, None] = "Source",
            container_type: Union[str, None] = "Snapshot-Full",
            snowflake: Union[str, None] = None,
            complexity: Union[str, None] = None,
            is_immutable: Union[bool, None] = None,
            is_mnpi: Union[bool, None] = None,
            is_pci: Union[bool, None] = None,
            is_pii: Union[bool, None] = None,
            is_client: Union[bool, None] = None,
            is_public: Union[bool, None] = None,
            is_internal: Union[bool, None] = None,
            is_confidential: Union[bool, None] = None,
            is_highly_confidential: Union[bool, None] = None,
            is_active: Union[bool, None] = None,
            owners: Union[List[str], None] = None,
            application_id: Union[str, Dict[str, str], None] = None,
            **kwargs: Any,
        ) -> Dataset:
            """Instantiate a Dataset object with this client for metadata creation."""
            dataset_obj = Dataset(
                identifier=identifier,
                title=title,
                category=category,
                description=description,
                frequency=frequency,
                is_internal_only_dataset=is_internal_only_dataset,
                is_third_party_data=is_third_party_data,
                is_restricted=is_restricted,
                is_raw_data=is_raw_data,
                maintainer=maintainer,
                source=source,
                region=region,
                publisher=publisher,
                product=product,
                sub_category=sub_category,
                tags=tags,
                created_date=created_date,
                modified_date=modified_date,
                delivery_channel=delivery_channel,
                language=language,
                status=status,
                type_=type_,
                container_type=container_type,
                snowflake=snowflake,
                complexity=complexity,
                is_immutable=is_immutable,
                is_mnpi=is_mnpi,
                is_pci=is_pci,
                is_pii=is_pii,
                is_client=is_client,
                is_public=is_public,
                is_internal=is_internal,
                is_confidential=is_confidential,
                is_highly_confidential=is_highly_confidential,
                is_active=is_active,
                owners=owners,
                application_id=application_id,
                **kwargs,
            )
            dataset_obj.client = self
            return dataset_obj


    def attribute(  # noqa: PLR0913
            self,
            identifier: str,
            index: int,
            data_type: Union[str, Types] = "String",
            title: str = "",
            description: str = "",
            is_dataset_key: bool = False,
            source: Optional[str] = None,
            source_field_id: Optional[str] = None,
            is_internal_dataset_key: Optional[bool] = None,
            is_externally_visible: Optional[bool] = True,
            unit: Optional[Any] = None,
            multiplier: float = 1.0,
            is_propagation_eligible: Optional[bool] = None,
            is_metric: Optional[bool] = None,
            available_from: Optional[str] = None,
            deprecated_from: Optional[str] = None,
            term: str = "bizterm1",
            dataset: Optional[int] = None,
            attribute_type: Optional[str] = None,
            application_id: Optional[Union[str, dict]] = None,
            **kwargs: Any,
        ) -> Attribute:
        data_type = Types[str(data_type).strip().rsplit(".", maxsplit=1)[-1].title()]
        attribute_obj = Attribute(
            identifier=identifier,
            index=index,
            data_type=data_type,
            title=title,
            description=description,
            is_dataset_key=is_dataset_key,
            source=source,
            source_field_id=source_field_id,
            is_internal_dataset_key=is_internal_dataset_key,
            is_externally_visible=is_externally_visible,
            unit=unit,
            multiplier=multiplier,
            is_propagation_eligible=is_propagation_eligible,
            is_metric=is_metric,
            available_from=available_from,
            deprecated_from=deprecated_from,
            term=term,
            dataset=dataset,
            attribute_type=attribute_type,
            application_id=application_id,
            **kwargs,
        )
        attribute_obj.client = self
        return attribute_obj

    def attributes(
            self,
            attributes: Optional[List[Attribute]] = None,
        ) -> Attributes:
        attributes_obj = Attributes(attributes=attributes or [])
        attributes_obj.client = self
        return attributes_obj

    def delete_datasetmembers(
            self,
            dataset: str,
            series_members: Union[str, List[str]],
            catalog: Optional[str] = None,
            return_resp_obj: bool = False,
        ) -> Optional[List[requests.Response]]:
        catalog = self._use_catalog(catalog)
        if isinstance(series_members, str):
            series_members = [series_members]
        responses = []
        for series_member in series_members:
            url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/datasetseries/{series_member}"
            resp = self.session.delete(url)
            requests_raise_for_status(resp)
            responses.append(resp)
        return responses if return_resp_obj else None

    def delete_all_datasetmembers(
            self,
            dataset: str,
            catalog: Optional[str] = None,
            return_resp_obj: bool = False,
        ) -> Optional[requests.Response]:
        catalog = self._use_catalog(catalog)
        url = f"{self.root_url}catalogs/{catalog}/datasets/{dataset}/datasetseries"
        resp = self.session.delete(url)
        requests_raise_for_status(resp)
        return resp if return_resp_obj else None

    def list_registered_attributes(
            self,
            catalog: Optional[str] = None,
            output: bool = False,
            display_all_columns: bool = False,
        ) -> pd.DataFrame:
        catalog = self._use_catalog(catalog)
        url = f"{self.root_url}catalogs/{catalog}/attributes"
        ds_attr_df = Fusion._call_for_dataframe(url, self.session).reset_index(drop=True)
        if not display_all_columns:
            ds_attr_df = ds_attr_df[
                ds_attr_df.columns.intersection(
                    [
                        "identifier",
                        "title",
                        "dataType",
                        "description",
                        "publisher",
                        "applicationId",
                    ]
                )
            ]
        if output:
            pass
        return ds_attr_df

    def report(  # noqa: PLR0913
            self,
            identifier: str,
            title: str = "",
            category: Union[str, List[str], None] = None,
            description: str = "",
            frequency: str = "Once",
            is_internal_only_dataset: bool = False,
            is_third_party_data: bool = True,
            is_restricted: Optional[bool] = None,
            is_raw_data: bool = True,
            maintainer: Optional[str] = "J.P. Morgan Fusion",
            source: Union[str, List[str], None] = None,
            region: Union[str, List[str], None] = None,
            publisher: str = "J.P. Morgan",
            product: Union[str, List[str], None] = None,
            sub_category: Union[str, List[str], None] = None,
            tags: Union[str, List[str], None] = None,
            created_date: Optional[str] = None,
            modified_date: Optional[str] = None,
            delivery_channel: Union[str, List[str]] = "API",
            language: str = "English",
            status: str = "Available",
            type_: Optional[str] = "Report",
            container_type: Optional[str] = "Snapshot-Full",
            snowflake: Optional[str] = None,
            complexity: Optional[str] = None,
            is_immutable: Optional[bool] = None,
            is_mnpi: Optional[bool] = None,
            is_pci: Optional[bool] = None,
            is_pii: Optional[bool] = None,
            is_client: Optional[bool] = None,
            is_public: Optional[bool] = None,
            is_internal: Optional[bool] = None,
            is_confidential: Optional[bool] = None,
            is_highly_confidential: Optional[bool] = None,
            is_active: Optional[bool] = None,
            owners: Optional[List[str]] = None,
            application_id: Union[str, Dict[str, str], None] = None,
            report: Optional[Dict[str, str]] = None,
            **kwargs: Any,
        ) -> Report:
            """Instantiate Report object with this client for metadata creation
            for managing regulatory reporting metadata.

            Args:
                identifier (str): Dataset identifier.
                title (str, optional): Dataset title. If not provided, defaults to identifier.
                category (Union[str, List[str], None], optional): A category or list of categories for the dataset.
                    Defaults to None.
                description (str, optional): Dataset description. If not provided, defaults to identifier.
                frequency (str, optional): The frequency of the dataset. Defaults to "Once".
                is_internal_only_dataset (bool, optional): Flag for internal datasets. Defaults to False.
                is_third_party_data (bool, optional): Flag for third party data. Defaults to True.
                is_restricted (Optional[bool], optional): Flag for restricted datasets. Defaults to None.
                is_raw_data (bool, optional): Flag for raw datasets. Defaults to True.
                maintainer (Optional[str], optional): Dataset maintainer. Defaults to "J.P. Morgan Fusion".
                source (Union[str, List[str], None], optional): Name of data vendor which provided the data.
                    Defaults to None.
                region (Union[str, List[str], None], optional): Region. Defaults to None.
                publisher (str, optional): Name of vendor that publishes the data. Defaults to "J.P. Morgan".
                product (Union[str, List[str], None], optional): Product to associate dataset with. Defaults to None.
                sub_category (Union[str, List[str], None], optional): Sub-category. Defaults to None.
                tags (Union[str, List[str], None], optional): Tags used for search purposes. Defaults to None.
                created_date (Optional[str], optional): Created date. Defaults to None.
                modified_date (Optional[str], optional): Modified date. Defaults to None.
                delivery_channel (Union[str, List[str]], optional): Delivery channel. Defaults to "API".
                language (str, optional): Language. Defaults to "English".
                status (str, optional): Status. Defaults to "Available".
                type_ (Optional[str], optional): Dataset type. Defaults to "Source".
                container_type (Optional[str], optional): Container type. Defaults to "Snapshot-Full".
                snowflake (Optional[str], optional): Snowflake account connection. Defaults to None.
                complexity (Optional[str], optional): Complexity. Defaults to None.
                is_immutable (Optional[bool], optional): Flag for immutable datasets. Defaults to None.
                is_mnpi (Optional[bool], optional): is_mnpi. Defaults to None.
                is_pci (Optional[bool], optional): is_pci. Defaults to None.
                is_pii (Optional[bool], optional): is_pii. Defaults to None.
                is_client (Optional[bool], optional): is_client. Defaults to None.
                is_public (Optional[bool], optional): is_public. Defaults to None.
                is_internal (Optional[bool], optional): is_internal. Defaults to None.
                is_confidential (Optional[bool], optional): is_confidential. Defaults to None.
                is_highly_confidential (Optional[bool], optional): is_highly_confidential. Defaults to None.
                is_active (Optional[bool], optional): is_active. Defaults to None.
                owners (Optional[List[str]], optional): The owners of the dataset. Defaults to None.
                application_id (Union[str, Dict[str, str], None], optional): The application ID of the dataset.
                    Defaults to None.
                report (Optional[Dict[str, str]], optional): The report metadata. Specifies the tier of the report.
                    Required for registered reports to the catalog.

            Returns:
                Report: Fusion Report class.

            Examples:
                >>> from fusion37 import Fusion
                >>> fusion = Fusion()
                >>> dataset = fusion.report(identifier="DATASET_1")

            Note:
                See the dataset module for more information on functionalities of report objects.

            """
            report_obj = Report(
                identifier=identifier,
                title=title,
                category=category,
                description=description,
                frequency=frequency,
                is_internal_only_dataset=is_internal_only_dataset,
                is_third_party_data=is_third_party_data,
                is_restricted=is_restricted,
                is_raw_data=is_raw_data,
                maintainer=maintainer,
                source=source,
                region=region,
                publisher=publisher,
                product=product,
                sub_category=sub_category,
                tags=tags,
                created_date=created_date,
                modified_date=modified_date,
                delivery_channel=delivery_channel,
                language=language,
                status=status,
                type_=type_,
                container_type=container_type,
                snowflake=snowflake,
                complexity=complexity,
                is_immutable=is_immutable,
                is_mnpi=is_mnpi,
                is_pci=is_pci,
                is_pii=is_pii,
                is_client=is_client,
                is_public=is_public,
                is_internal=is_internal,
                is_confidential=is_confidential,
                is_highly_confidential=is_highly_confidential,
                is_active=is_active,
                owners=owners,
                application_id=application_id,
                report=report,
                **kwargs,
            )
            report_obj.client = self
            return report_obj

    def input_dataflow(  # noqa: PLR0913
        self,
        identifier: str,
        title: str = "",
        category: Union[str, List[str], None] = None,
        description: str = "",
        frequency: str = "Once",
        is_internal_only_dataset: bool = False,
        is_third_party_data: bool = True,
        is_restricted: Optional[bool] = None,
        is_raw_data: bool = True,
        maintainer: Optional[str] = "J.P. Morgan Fusion",
        source: Union[str, List[str], None] = None,
        region: Union[str, List[str], None] = None,
        publisher: str = "J.P. Morgan",
        product: Union[str, List[str], None] = None,
        sub_category: Union[str, List[str], None] = None,
        tags: Union[str, List[str], None] = None,
        created_date: Optional[str] = None,
        modified_date: Optional[str] = None,
        delivery_channel: Union[str, List[str]] = "API",
        language: str = "English",
        status: str = "Available",
        type_: Optional[str] = "Flow",
        container_type: Optional[str] = "Snapshot-Full",
        snowflake: Optional[str] = None,
        complexity: Optional[str] = None,
        is_immutable: Optional[bool] = None,
        is_mnpi: Optional[bool] = None,
        is_pci: Optional[bool] = None,
        is_pii: Optional[bool] = None,
        is_client: Optional[bool] = None,
        is_public: Optional[bool] = None,
        is_internal: Optional[bool] = None,
        is_confidential: Optional[bool] = None,
        is_highly_confidential: Optional[bool] = None,
        is_active: Optional[bool] = None,
        owners: Optional[List[str]] = None,
        application_id: Union[str, Dict[str, str], None] = None,
        producer_application_id: Optional[Dict[str, str]] = None,
        consumer_application_id: Union[List[Dict[str, str]], Dict[str, str], None] = None,
        flow_details: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> InputDataFlow:
        flow_details = {"flowDirection": "Input"} if flow_details is None else flow_details
        dataflow_obj = InputDataFlow(
            identifier=identifier,
            title=title,
            category=category,
            description=description,
            frequency=frequency,
            is_internal_only_dataset=is_internal_only_dataset,
            is_third_party_data=is_third_party_data,
            is_restricted=is_restricted,
            is_raw_data=is_raw_data,
            maintainer=maintainer,
            source=source,
            region=region,
            publisher=publisher,
            product=product,
            sub_category=sub_category,
            tags=tags,
            created_date=created_date,
            modified_date=modified_date,
            delivery_channel=delivery_channel,
            language=language,
            status=status,
            type_=type_,
            container_type=container_type,
            snowflake=snowflake,
            complexity=complexity,
            is_immutable=is_immutable,
            is_mnpi=is_mnpi,
            is_pci=is_pci,
            is_pii=is_pii,
            is_client=is_client,
            is_public=is_public,
            is_internal=is_internal,
            is_confidential=is_confidential,
            is_highly_confidential=is_highly_confidential,
            is_active=is_active,
            owners=owners,
            application_id=application_id,
            producer_application_id=producer_application_id,
            consumer_application_id=consumer_application_id,
            flow_details=flow_details,
            **kwargs,
        )
        dataflow_obj.client = self
        return dataflow_obj

    def output_dataflow(  # noqa: PLR0913
        self,
        identifier: str,
        title: str = "",
        category: Union[str, List[str], None] = None,
        description: str = "",
        frequency: str = "Once",
        is_internal_only_dataset: bool = False,
        is_third_party_data: bool = True,
        is_restricted: Optional[bool] = None,
        is_raw_data: bool = True,
        maintainer: Optional[str] = "J.P. Morgan Fusion",
        source: Union[str, List[str], None] = None,
        region: Union[str, List[str], None] = None,
        publisher: str = "J.P. Morgan",
        product: Union[str, List[str], None] = None,
        sub_category: Union[str, List[str], None] = None,
        tags: Union[str, List[str], None] = None,
        created_date: Optional[str] = None,
        modified_date: Optional[str] = None,
        delivery_channel: Union[str, List[str]] = "API",
        language: str = "English",
        status: str = "Available",
        type_: Optional[str] = "Flow",
        container_type: Optional[str] = "Snapshot-Full",
        snowflake: Optional[str] = None,
        complexity: Optional[str] = None,
        is_immutable: Optional[bool] = None,
        is_mnpi: Optional[bool] = None,
        is_pci: Optional[bool] = None,
        is_pii: Optional[bool] = None,
        is_client: Optional[bool] = None,
        is_public: Optional[bool] = None,
        is_internal: Optional[bool] = None,
        is_confidential: Optional[bool] = None,
        is_highly_confidential: Optional[bool] = None,
        is_active: Optional[bool] = None,
        owners: Optional[List[str]] = None,
        application_id: Union[str, Dict[str, str], None] = None,
        producer_application_id: Optional[Dict[str, str]] = None,
        consumer_application_id: Union[List[Dict[str, str]], Dict[str, str], None] = None,
        flow_details: Optional[Dict[str, str]] = None,
        **kwargs: Any,
    ) -> OutputDataFlow:
        flow_details = {"flowDirection": "Output"} if flow_details is None else flow_details
        dataflow_obj = OutputDataFlow(
            identifier=identifier,
            title=title,
            category=category,
            description=description,
            frequency=frequency,
            is_internal_only_dataset=is_internal_only_dataset,
            is_third_party_data=is_third_party_data,
            is_restricted=is_restricted,
            is_raw_data=is_raw_data,
            maintainer=maintainer,
            source=source,
            region=region,
            publisher=publisher,
            product=product,
            sub_category=sub_category,
            tags=tags,
            created_date=created_date,
            modified_date=modified_date,
            delivery_channel=delivery_channel,
            language=language,
            status=status,
            type_=type_,
            container_type=container_type,
            snowflake=snowflake,
            complexity=complexity,
            is_immutable=is_immutable,
            is_mnpi=is_mnpi,
            is_pci=is_pci,
            is_pii=is_pii,
            is_client=is_client,
            is_public=is_public,
            is_internal=is_internal,
            is_confidential=is_confidential,
            is_highly_confidential=is_highly_confidential,
            is_active=is_active,
            owners=owners,
            application_id=application_id,
            producer_application_id=producer_application_id,
            consumer_application_id=consumer_application_id,
            flow_details=flow_details,
            **kwargs,
        )
        dataflow_obj.client = self
        return dataflow_obj

    def to_df(  # noqa: PLR0913
            self,
            dataset: str,
            dt_str: str = "latest",
            dataset_format: str = "parquet",
            catalog: str | None = None,
            n_par: int | None = None,
            show_progress: bool = True,
            columns: list[str] | None = None,
            filters: Any | None = None,
            force_download: bool = False,
            download_folder: str | None = None,
            dataframe_type: str = "pandas",
            **kwargs: Any,
    ) -> pd.DataFrame:
        """Gets distributions for a specified date or date range and returns the data as a dataframe.

        Args:
            dataset (str): A dataset identifier
            dt_str (str, optional): Either a single date or a range identified by a start or end date,
                or both separated with a ":". Defaults to 'latest' which will return the most recent
                instance of the dataset.
            dataset_format (str, optional): The file format, e.g. CSV or Parquet. Defaults to 'parquet'.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            n_par (int, optional): Specify how many distributions to download in parallel.
                Defaults to all cpus available.
            show_progress (bool, optional): Display a progress bar during data download Defaults to True.
            columns (List, optional): A list of columns to return from a parquet file. Defaults to None
            filters (List, optional): List[Tuple] or List[List[Tuple]] or None (default)
                Rows which do not match the filter predicate will be removed from scanned data.
                Partition keys embedded in a nested directory structure will be exploited to avoid
                loading files at all if they contain no matching rows. If use_legacy_dataset is True,
                filters can only reference partition keys and only a hive-style directory structure
                is supported. When setting use_legacy_dataset to False, also within-file level filtering
                and different partitioning schemes are supported.
                More on https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
            force_download (bool, optional): If True then will always download a file even
                if it is already on disk. Defaults to False.
            download_folder (str, optional): The path, absolute or relative, where downloaded files are saved.
                Defaults to download_folder as set in __init__
            dataframe_type (str, optional): Type
        Returns:
            class:`pandas.DataFrame`: a dataframe containing the requested data.
                If multiple dataset instances are retrieved then these are concatenated first.
        """
        raise NotImplementedError("Method not implemented")

    def to_table(  # noqa: PLR0913
            self,
            dataset: str,
            dt_str: str = "latest",
            dataset_format: str = "parquet",
            catalog: str | None = None,
            n_par: int | None = None,
            show_progress: bool = True,
            columns: list[str] | None = None,
            filters: Any | None = None,
            force_download: bool = False,
            download_folder: str | None = None,
            **kwargs: Any,
    ) -> Any:
        """Gets distributions for a specified date or date range and returns the data as an arrow table.

        Args:
            dataset (str): A dataset identifier
            dt_str (str, optional): Either a single date or a range identified by a start or end date,
                or both separated with a ":". Defaults to 'latest' which will return the most recent
                instance of the dataset.
            dataset_format (str, optional): The file format, e.g. CSV or Parquet. Defaults to 'parquet'.
            catalog (str, optional): A catalog identifier. Defaults to 'common'.
            n_par (int, optional): Specify how many distributions to download in parallel.
                Defaults to all cpus available.
            show_progress (bool, optional): Display a progress bar during data download Defaults to True.
            columns (List, optional): A list of columns to return from a parquet file. Defaults to None
            filters (List, optional): List[Tuple] or List[List[Tuple]] or None (default)
                Rows which do not match the filter predicate will be removed from scanned data.
                Partition keys embedded in a nested directory structure will be exploited to avoid
                loading files at all if they contain no matching rows. If use_legacy_dataset is True,
                filters can only reference partition keys and only a hive-style directory structure
                is supported. When setting use_legacy_dataset to False, also within-file level filtering
                and different partitioning schemes are supported.
                More on https://arrow.apache.org/docs/python/generated/pyarrow.parquet.ParquetDataset.html
            force_download (bool, optional): If True then will always download a file even
                if it is already on disk. Defaults to False.
            download_folder (str, optional): The path, absolute or relative, where downloaded files are saved.
                Defaults to download_folder as set in __init__
        Returns:
            class:`pyarrow.Table`: a dataframe containing the requested data.
                If multiple dataset instances are retrieved then these are concatenated first.
        """
        raise NotImplementedError("Method not implemented")

    def listen_to_events(
            self,
            last_event_id: str | None = None,
            catalog: str | None = None,
            url: str = "https://fusion.jpmorgan.com/api/v1/",
    ) -> None | pd.DataFrame:
        """Run server sent event listener in the background. Retrieve results by running get_events.

        Args:
            last_event_id (str): Last event ID (exclusive).
            catalog (str): catalog.
            url (str): subscription url.
        Returns:
            Union[None, class:`pandas.DataFrame`]: If in_background is True then the function returns no output.
                If in_background is set to False then pandas DataFrame is output upon keyboard termination.
        """
        raise NotImplementedError("Method not implemented")

    def get_events(
            self,
            last_event_id: str | None = None,
            catalog: str | None = None,
            in_background: bool = True,
            url: str = "https://fusion.jpmorgan.com/api/v1/",
    ) -> None | pd.DataFrame:
        """Run server sent event listener and print out the new events. Keyboard terminate to stop.

        Args:
            last_event_id (str): id of the last event.
            catalog (str): catalog.
            in_background (bool): execute event monitoring in the background (default = True).
            url (str): subscription url.
        Returns:
            Union[None, class:`pandas.DataFrame`]: If in_background is True then the function returns no output.
                If in_background is set to False then pandas DataFrame is output upon keyboard termination.
        """
        raise NotImplementedError("Method not implemented")


"""Fusion Product class and functions."""

from __future__ import annotations

import json as js
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

import pandas as pd

from .utils import (
    CamelCaseMeta,
    _is_json,
    camel_to_snake,
    convert_date_format,
    make_bool,
    make_list,
    requests_raise_for_status,
    snake_to_camel,
    tidy_string,
)

if TYPE_CHECKING:
    import requests

    from .fusion import Fusion


@dataclass
class Product(metaclass=CamelCaseMeta):
    """Fusion Product class for managing product metadata in a Fusion catalog.

    Attributes:
        identifier (str): A unique identifier for the product.
        title (str, optional): Product title. Defaults to "".
        category (Union[str, List[str], None], optional): Product category. Defaults to None.
        short_abstract (str, optional): Short abstract of the product. Defaults to "".
        description (str, optional): Product description. If not provided, defaults to identifier.
        is_active (bool, optional): Boolean for Active status. Defaults to True.
        is_restricted (Optional[bool], optional): Flag for restricted products. Defaults to None.
        maintainer (Union[str, List[str], None], optional): Product maintainer. Defaults to None.
        region (Union[str, List[str]], optional): Product region. Defaults to ["Global"].
        publisher (Optional[str], optional): Name of vendor that publishes the data. Defaults to None.
        sub_category (Union[str, List[str], None], optional): Product sub-category. Defaults to None.
        tag (Union[str, List[str], None], optional): Tags used for search purposes. Defaults to None.
        delivery_channel (Union[str, List[str]], optional): Product delivery channel. Defaults to ["API"].
        theme (Optional[str], optional): Product theme. Defaults to None.
        release_date (Optional[str], optional): Product release date. Defaults to None.
        language (str, optional): Product language. Defaults to "English".
        status (str, optional): Product status. Defaults to "Available".
        image (str, optional): Product image. Defaults to "".
        logo (str, optional): Product logo. Defaults to "".
        dataset (Union[str, List[str], None], optional): Product datasets. Defaults to None.
        _client (Any, optional): Fusion client object. Defaults to None.

    """

    identifier: str
    title: str = ""
    category: Union[str, List[str], None] = None
    short_abstract: str = ""
    description: str = ""
    is_active: bool = True
    is_restricted: Optional[bool] = None
    maintainer: Union[str, List[str], None] = None
    region: Union[str, List[str]] = field(default_factory=lambda: ["Global"])
    publisher: str = "J.P. Morgan"
    sub_category: Union[str, List[str], None] = None
    tag: Union[str, List[str], None] = None
    delivery_channel: Union[str, List[str]] = field(default_factory=lambda: ["API"])
    theme: Optional[str] = None
    release_date: Optional[str] = None
    language: str = "English"
    status: str = "Available"
    image: str = ""
    logo: str = ""
    dataset: Union[str, List[str], None] = None

    _client: Optional[Any] = field(init=False, repr=False, compare=False, default=None)

    def __repr__(self) -> str:
        """Return an object representation of the Product object.

        Returns:
            str: Object representation of the product.

        """
        attrs = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return f"Product(\n" + ",\n ".join(f"{k}={v!r}" for k, v in attrs.items()) + "\n)"

    def __post_init__(self) -> None:
        """Format Product metadata fields after object instantiation."""
        self.identifier = tidy_string(self.identifier).upper().replace(" ", "_")
        self.title = tidy_string(self.title) if self.title != "" else self.identifier.replace("_", " ").title()
        self.description = tidy_string(self.description) if self.description != "" else self.title
        self.short_abstract = tidy_string(self.short_abstract) if self.short_abstract != "" else self.title
        self.description = tidy_string(self.description)
        self.category = (
            self.category if isinstance(self.category, list) or self.category is None else make_list(self.category)
        )
        self.tag = self.tag if isinstance(self.tag, list) or self.tag is None else make_list(self.tag)
        self.dataset = (
            self.dataset if isinstance(self.dataset, list) or self.dataset is None else make_list(self.dataset)
        )
        self.sub_category = (
            self.sub_category
            if isinstance(self.sub_category, list) or self.sub_category is None
            else make_list(self.sub_category)
        )
        self.is_active = self.is_active if isinstance(self.is_active, bool) else make_bool(self.is_active)
        self.is_restricted = (
            self.is_restricted
            if isinstance(self.is_restricted, bool) or self.is_restricted is None
            else make_bool(self.is_restricted)
        )
        self.maintainer = (
            self.maintainer
            if isinstance(self.maintainer, list) or self.maintainer is None
            else make_list(self.maintainer)
        )
        self.region = self.region if isinstance(self.region, list) or self.region is None else make_list(self.region)
        self.delivery_channel = (
            self.delivery_channel if isinstance(self.delivery_channel, list) else make_list(self.delivery_channel)
        )
        self.release_date = convert_date_format(self.release_date) if self.release_date else None

    def __getattr__(self, name: str) -> Any:
        # Redirect attribute access to the snake_case version
        snake_name = camel_to_snake(name)
        if snake_name in self.__dict__:
            return self.__dict__[snake_name]
        raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "client":
            # Use the property setter for client
            object.__setattr__(self, name, value)
        else:
            snake_name = camel_to_snake(name)
            self.__dict__[snake_name] = value

    
    @property
    def client(self) -> Optional[Any]:
        """Return the client."""
        return self._client


    @client.setter
    def client(self, client: Optional[Any]) -> None:
        """Set the client for the Product. Set automatically, if the Product is instantiated from a Fusion object.

        Args:
            client (Any): Fusion client object.

        Examples:
            >>> from fusion37 import Fusion
            >>> fusion = Fusion()
            >>> product = fusion.product("my_product")
            >>> product.client = fusion

        """
        self._client = client


    def _use_client(self, client: Optional[Any]) -> Any:
        """Determine client."""
        res = self._client if client is None else client
        if res is None:
            raise ValueError("A Fusion client object is required.")
        return res


    @classmethod
    def _from_series(cls: Type[Product], series: pd.Series) -> Product:
        """Instantiate a Product object from a pandas Series.

        Args:
            series (pd.Series): Product metadata as a pandas Series.

        Returns:
            Product: Product object.

        """
        series = series.rename(lambda x: x.replace(" ", "").replace("_", "").lower())
        series = series.rename({"tag": "tags", "dataset": "datasets"})
        short_abstract = series.get("abstract", "")
        short_abstract = series.get("shortabstract", "") if short_abstract is None else short_abstract

        return cls(
            title=series.get("title", ""),
            identifier=series.get("identifier", ""),
            category=series.get("category", None),
            short_abstract=short_abstract,
            description=series.get("description", ""),
            theme=series.get("theme", None),
            release_date=series.get("releasedate", None),
            is_active=series.get("isactive", True),
            is_restricted=series.get("isrestricted", None),
            maintainer=series.get("maintainer", None),
            region=series.get("region", "Global"),
            publisher=series.get("publisher", "J.P. Morgan"),
            sub_category=series.get("subcategory", None),
            tag=series.get("tags", None),
            delivery_channel=series.get("deliverychannel", "API"),
            language=series.get("language", "English"),
            status=series.get("status", "Available"),
            dataset=series.get("datasets", None),
        )


    @classmethod
    def _from_dict(cls: Type[Product], data: dict) -> Product:
        """Instantiate a Product object from a dictionary.

        Args:
            data (dict): Product metadata as a dictionary.

        Returns:
            Product: Product object.

        """
        keys = [f.name for f in fields(cls)]
        data = {camel_to_snake(k): v for k, v in data.items()}
        data = {k: v for k, v in data.items() if k in keys}
        return cls(**data)


    @classmethod
    def _from_csv(cls: Type[Product], file_path: str, identifier: Optional[str] = None) -> Product:
        """Instantiate a Product object from a CSV file.

        Args:
            file_path (str): Path to the CSV file.
            identifier (Optional[str], optional): Product identifier for filtering if multiple 
            products are defined in CSV. Defaults to None.

        Returns:
            Product: Product object.

        """
        data = pd.read_csv(file_path)

        return (
            Product._from_series(data[data["identifier"] == identifier].reset_index(drop=True).iloc[0])
            if identifier
            else Product._from_series(data.reset_index(drop=True).iloc[0])
        )

    def from_object(
        self,
        product_source: Union[Product, Dict[str, Any], str, pd.Series]
    ) -> Product:
        """Instantiate a Product object from a Product object, dictionary, path to CSV, JSON string, or pandas Series.

        Args:
            product_source (Union[Product, dict[str, Any], str, pd.Series]): Product metadata source.

        Raises:
            TypeError: If the object provided is not a Product, dictionary, path to CSV file, JSON string,
            or pandas Series.

        Returns:
            Product: Product object.

        Examples:
            Instantiating a Product object from a dictionary:

            >>> from fusion37 import Fusion
            >>> from fusion37.product import Product
            >>> fusion = Fusion()
            >>> product_dict = { ... }
            >>> product = fusion.product("my_product").from_object(product_dict)
        """
        if isinstance(product_source, Product):
            product = product_source
        elif isinstance(product_source, dict):
            product = Product._from_dict(product_source)
        elif isinstance(product_source, str):
            if _is_json(product_source):
                product = Product._from_dict(js.loads(product_source))
            else:
                product = Product._from_csv(product_source)
        elif isinstance(product_source, pd.Series):
            product = Product._from_series(product_source)
        else:
            raise TypeError(f"Could not resolve the object provided: {product_source}")
        product.client = self._client
        return product


    def from_catalog(self, catalog: Optional[str] = None, client: Optional[Fusion] = None) -> Product:
        """Instantiate a Product object from a Fusion catalog.

        Args:
            catalog (Optional[str], optional): Catalog identifier. Defaults to None.
            client (Optional[Fusion], optional): Fusion session. Defaults to None.

        Returns:
            Product: Product object.
        """
        client = self._use_client(client)
        catalog = client._use_catalog(catalog)

        resp = client.session.get(f"{client.root_url}catalogs/{catalog}/products")
        requests_raise_for_status(resp)
        list_products = resp.json()["resources"]
        filtered_products = [dict_ for dict_ in list_products if dict_["identifier"] == self.identifier]
        if not filtered_products:
            raise ValueError(f"Product with identifier '{self.identifier}' not found in catalog '{catalog}'.")
        dict_ = filtered_products[0]
        product_obj = Product._from_dict(dict_)
        product_obj.client = client

        return product_obj


    def to_dict(self: Product) -> Dict[str, Any]:
        """Convert the Product instance to a dictionary.

        Returns:
            Dict[str, Any]: Product metadata as a dictionary.
        """
        product_dict = {
            snake_to_camel(k): v
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }
        return product_dict


    def create(
        self,
        catalog: Optional[str] = None,
        client: Optional[Fusion] = None,
        return_resp_obj: bool = False,
    ) -> Optional[requests.Response]:
        """Upload a new product to a Fusion catalog.

        Args:
            client (Optional[Fusion], optional): A Fusion client object. Defaults to the instance's _client.
            catalog (Optional[str], optional): A catalog identifier. Defaults to None.
            return_resp_obj (bool, optional): If True then return the response object. Defaults to False.

        Returns:
            Optional[requests.Response]: The response object from the API call if return_resp_obj is True, 
            otherwise None.
        """
        client = self._use_client(client)
        catalog = client._use_catalog(catalog)

        release_date = self.release_date if self.release_date else pd.Timestamp("today").strftime("%Y-%m-%d")
        delivery_channel = self.delivery_channel if self.delivery_channel else ["API"]

        self.release_date = release_date
        self.delivery_channel = delivery_channel

        data = self.to_dict()

        url = f"{client.root_url}catalogs/{catalog}/products/{self.identifier}"
        resp: requests.Response = client.session.post(url, json=data)
        requests_raise_for_status(resp)
        return resp if return_resp_obj else None

    def update(
        self,
        catalog: Optional[str] = None,
        client: Optional[Fusion] = None,
        return_resp_obj: bool = False,
    ) -> Optional[requests.Response]:
        """Update an existing product in a Fusion catalog.

        Args:
            client (Fusion): A Fusion client object. Defaults to the instance's 
                _client.
            catalog (Optional[str], optional): A catalog identifier. Defaults to 
                None.
            return_resp_obj (bool, optional): If True then return the response 
                object. Defaults to False.

        Returns:
            Optional[requests.Response]: The response object from the API call 
            if return_resp_obj is True, otherwise None.
        """
        client = self._use_client(client)
        catalog = client._use_catalog(catalog)

        release_date = self.release_date if self.release_date else pd.Timestamp("today").strftime("%Y-%m-%d")
        delivery_channel = self.delivery_channel if self.delivery_channel else ["API"]

        self.release_date = release_date
        self.delivery_channel = delivery_channel

        data = self.to_dict()

        url = f"{client.root_url}catalogs/{catalog}/products/{self.identifier}"
        resp: requests.Response = client.session.put(url, json=data)
        requests_raise_for_status(resp)
        return resp if return_resp_obj else None


    def delete(
        self,
        catalog: Optional[str] = None,
        client: Optional[Fusion] = None,
        return_resp_obj: bool = False,
    ) -> Optional[requests.Response]:
        """Delete a product from a Fusion catalog.

        Args:
            client (Fusion): A Fusion client object. Defaults to the instance's _client.
            catalog (Optional[str], optional): A catalog identifier. Defaults to None.
            return_resp_obj (bool, optional): If True then return the response object. Defaults to False.

        Returns:
            Optional[requests.Response]: The response object from the API call 
            if return_resp_obj is True, otherwise None.
        """
        client = self._use_client(client)
        catalog = client._use_catalog(catalog)

        url = f"{client.root_url}catalogs/{catalog}/products/{self.identifier}"
        resp: requests.Response = client.session.delete(url)
        requests_raise_for_status(resp)
        return resp if return_resp_obj else None


    def copy(
        self,
        catalog_to: str,
        catalog_from: Optional[str] = None,
        client: Optional[Fusion] = None,
        client_to: Optional[Fusion] = None,
        return_resp_obj: bool = False,
    ) -> Optional[requests.Response]:
        """Copy product from one Fusion catalog and/or environment to another by copy.

        Args:
            catalog_to (str): Catalog identifier to which to copy product.
            catalog_from (Optional[str], optional): A catalog identifier from which to copy product. 
            Defaults to "common".
            client (Fusion): A Fusion client object. Defaults to the instance's _client.
            client_to (Optional[Fusion], optional): Fusion client object. Defaults to current instance.
            return_resp_obj (bool, optional): If True then return the response object. Defaults to False.

        Returns:
            Optional[requests.Response]: The response object from the API call if return_resp_obj is True, 
            otherwise None.
        """
        client = self._use_client(client)
        catalog_from = client._use_catalog(catalog_from)
        if client_to is None:
            client_to = client
        product_obj = self.from_catalog(catalog=catalog_from, client=client)
        product_obj.client = client_to
        resp = product_obj.create(catalog=catalog_to, return_resp_obj=True)
        return resp if return_resp_obj else None

